//  CurvedSpaces-Complete-D3D.cpp
//
//  This sample code accompanies the article
//
//      "Real-time rendering in curved spaces",
//      IEEE Computer Graphics and Applications, Nov-Dec 2002.
//
//  Interactive 3D animations in spherical, flat, and hyperbolic
//  spaces run in hardware on the user's 3D graphics card via Direct3D.
//
//  Different versions of this code serve different purposes.
//
//      CurvedSpaces-Simple-[D3D or GL].cpp
//
//          This stripped-down version of the program introduces
//          basic techniques of rendering in curved spaces.
//
//      CurvedSpaces-Complete-[D3D or GL].cpp
//
//          This more complete version of the program lets
//          the user explore a variety of 3-dimensional spaces,
//          and may be of interest to geometers and cosmologists
//          as well as computer graphics developers.
//
//  Readers may wish to run the richer CurvedSpaces-Complete.exe
//  to get a feel for what is possible, but read the source code
//  for the more modest CurvedSpaces-Simple.cpp to understand
//  the basics of the implementation.
//
//  Jeff Weeks
//  weeks@geometrygames.org
//  March 2004

//  This code requires the import libraries
//
//    d3dx.lib  ddraw.lib  user32.lib  gdi32.lib  comdlg32.lib 
//
//  For a debug build, link to d3dxd.lib instead of d3dx.lib.

//  Known issues:
//
//  (1) Direct3D cannot render into the red and blue-green channels
//      independently, so there is no clean way to implement
//      stereoscopic 3D.  The present code uses a workaround:
//      it sorts the scenery front-to-back, and then draws the
//      right-eye view over the left-eye view, using
//      D3DRENDERSTATE_SRCBLEND = D3DBLEND_ONE and
//      D3DRENDERSTATE_DESTBLEND = D3DBLEND_ONE.  This technique
//      yields a more-or-less correct image, except that the
//      beam surfaces have bright spots where they overlap
//      at junctions.  If a future version of Direct3D permits
//      masking the red, green, and blue channels independently,
//      then we can omit the sorting code and disable blending.
//      Note:  The OpenGL version of this program uses glColorMask()
//      to write to the red, green, and blue channels independently;
//      the trade-off is that OpenGL does not allow explicit
//      fog values, but offers only preset flat-space formulas.
//
//  (2) Direct3D and OpenGL use different coordinate systems.
//      By default Direct3D looks in the positive z-direction while
//      OpenGL looks in the negative z-direction.  Thus the scene
//      viewed in one will be the mirror image of the scene viewed
//      in the other.
//
//  (3) Direct3D "vertex buffers" might speed the code up a bit
//      by putting the spinning Earth and maybe also the Dirichlet
//      domain beams onto the graphics card, but I haven't done this.
//      (In the OpenGL version using "display lists" for the
//      spinning Earth helped but doing the same for the beams
//      had little effect.)

#include "Macros-D3D.h"
#include "DirichletDomain-D3D.hpp"
#include "SpinningEarth-D3D.hpp"
#include "CurvedSpacesResourceIDs-Complete-D3D.h"
#include <direct.h> //  for _getcwd()
#include <stdio.h>  //  for file I/O

//  How many pixels of empty screen space should
//  we allow around the default window position?
#define WINDOW_MARGIN   16

//  What is the minimum acceptable size for the
//  window's client area?
#define MIN_CLIENT_SIZE  192

//  Pressing the up-arrow key increases the camera speed
//  by CAMERA_SPEED_INCREMENT.  Pressing the down-arrow key
//  decreases it by that amount.  Pressing the space bar
//  sets the camera speed to zero.
#define CAMERA_SPEED_INCREMENT  0.02f

//  After clicking into the scene, moving the mouse pitches and yaws
//  the camera at a rate proportional to CAMERA_ROTATION_RATE.
//  Holding the Ctrl key down while moving the mouse rolls the camera.
#define CAMERA_ROTATION_RATE    1.0f

//  The CLIPPING_DISTANCE should be as close to the camera
//  as possible without inducing z-buffer related artifacts.
//  In a scene of depth ~1, a clipping distance of 0.005f
//  or greater works fine with an ordinary 16-bit depth buffer.
//  [Exercise for advanced readers:  Why would w-buffering fail
//  in spherical rendering?]
#define CLIPPING_DISTANCE   0.01f

//  How closely must the entries match for two matrices
//  to be considered the same?
#define MATRIX_EPSILON      1e-3

//  In a spherical space, objects near the antipodal point
//  appear just as large as nearby objects.  Fog provides
//  an essential depth cue, distinguishing distant large
//  images from nearby ones.  Fog also serves to limit
//  the potentially infinite view in flat and hyperbolic spaces.

//  The format of the FOG_COLOR is AARRGGBB, so for example
//  0xFF80FFFF is half-saturated cyan.
#define FOG_COLOR_MONO          0xFF000000
#define FOG_COLOR_STEREO        0xFF000000

//  The flat space fog parameters must be chosen consistently
//  with the typical distance scale of the scenery.
#define FLAT_FOG_BEGIN          1.5f
#define FLAT_FOG_END            2.0f

//  In hyperbolic and spherical spaces there is a natural unit
//  of distance (the radian) so the fog parameters may be set
//  in absolute terms, without reference to the scenery.
#define HYPERBOLIC_FOG_BEGIN    3.0f    //  1.0f
#define HYPERBOLIC_FOG_END      4.0f    //  2.0f
    
//  Fog in a spherical space requires special care.
//  The given settings spread the fog out over the front hemisphere,
//  and leave the back hemisphere rather dim.
#define SPHERICAL_FOG_AT_ANTIPODE   0.8f
#define SPHERICAL_FOG_MAX           0.9f

//  Precompute the Direct3D fog factor on the background image
//  in a spherical space.
#define SPHERICAL_BACKGROUND_FOG    ((((DWORD)((1.0f - SPHERICAL_FOG_MAX) * 0xFF)) << 24) | 0x00FFFFFF)

//  When testing for fixed points, how close must det(g - I)
//  be to zero to be considered singular?  Matrices with
//  fixed points have det(g - I) very small, so choose
//  SINGULAR_EPSILON large enough to avoid falsely rejecting
//  groups with short translation distances.
#define SINGULAR_EPSILON    1e-8

//  Mouse wheel support is available in Windows ME and later.
#ifndef WM_MOUSEWHEEL
#define WM_MOUSEWHEEL   0x020A
#endif

//  Direct3D specifies a color as ARGB, but Win32 uses ABGR.
#define RGB2BRG(c)      (                               \
                             (c) & 0xFF000000           \
                          | ((c) & 0x00FF0000) >> 16    \
                          |  (c) & 0x0000FF00           \
                          | ((c) & 0x000000FF) << 16    \
                        )


//  Basic curved space rendering proceeds without knowing
//  the curvature of space.  The exceptions are in
//
//  (1) PrependTranslation() which must know whether to create
//      a spherical, flat, or hyperbolic rotation matrix when
//      translating the camera or scenery position,
//
//  (2) NormalizeCameraPosition() which maintains an orthogonal
//      matrix for gCameraPosition, and
//
//  (3) Render() which in the spherical case must draw the back
//      hemisphere separately from the front hemisphere, as
//      explained in the accompanying article.
//
//  In CurvedSpacesSimple.cpp these are the *only* uses of
//  the SpaceType.  In CurvedSpacesComplete.cpp the SpaceType
//  is also taken into account when computing fog effects and
//  when constructing certain types of scenery.
typedef enum
{
    SpaceSpherical = 0,
    SpaceFlat,
    SpaceHyperbolic
} SpaceType;

//  In some flat and hyperbolic spaces it's possible to take
//  a trip and return as one's own mirror image, just as a
//  2-dimensional traveller on a Moebius strip returns as her
//  own mirror image.  In these cases we'll need to keep track
//  of whether the camera is mirror reversed or not.
typedef enum
{
    ParityNormal,
    ParityReversed
} ParityType;


//  The possible projection matrices are indexed according
//  to which eye is viewing the scene and whether we are mapping
//  the viewing banana into the front half of the clipping box
//  in projection space, into the back half, or into the whole box. 
typedef enum
{
    EyeOnly = 0,
    EyeLeft,
    EyeRight
} EyeType;
typedef enum
{
    BananaFull,
    BananaFront,
    BananaBack
} RangeType;


//  Different kinds of scenery will require different textures.
//  Be sure to keep this typedef synchronized with gTextureFiles below.
typedef enum
{
    TextureEarthNorthern = 0,
    TextureEarthSouthern,
    TextureBeam,
    TextureSpaceship,
    NumTextures
} TextureType;

//  Keep D3DXMATRICES on a linked list while producing the full
//  set of matrices from a set of generators.
typedef struct MatrixNode
{
    D3DXMATRIX          d3dm;
    struct MatrixNode   *next;
} MatrixNode;



//  Is fog enabled?
static BOOL gFog = TRUE;

//  Is stereoscopic 3D enabled?
//
//  Technical Note:  Stereoscopic 3D could be made rigorously correct
//  if we could render into the red and cyan channels independently,
//  as we do in the OpenGL version of this program.
//  Unfortunately D3D 7 isn't set up to do this.  It might be possible
//  to render the left and right eye views to separate DirectDraw surfaces
//  and then merge them, but I can't see how to do this and don't
//  presently have the time to figure it out.  For now we avoid most
//  artifacts by using blend factors of D3DBLEND_ONE and rendering
//  the scene front to back.  This gives an essentially correct picture
//  in the front hemisphere (except for some bright spots where the beams
//  meet), but fails to properly eclipse distant objects when
//  gRenderBackHemisphere is on.
static BOOL gStereo = FALSE;

//  In principle your line of sight in a hypersphere may reach across
//  the hemisphere in front of you, pass through your antipodal point,
//  and arrive at an object in the hemisphere behind you.  In practice
//  this won't happen if the scenery is symmetrically arranged,
//  with every scenery object paired with an identical, antipodal
//  scenery object.  In such cases we can omit the back hemisphere
//  and double the frame rate.
static BOOL gRenderBackHemisphere = FALSE;

//  Show the frame rate?
static BOOL gShowFrameRate = FALSE;

//  Is the current space spherical, flat, or hyperbolic?
static SpaceType gSpaceType = SpaceFlat;

//  Keep track of the camera's position, speed, and whether or not
//  it's mirror reversed.
static D3DXMATRIX       gCameraPosition     = D3DXMATRIX(
                                                1,0,0,0,
                                                0,1,0,0,
                                                0,0,1,0,
                                                0,0,0,1);
static float            gCameraSpeed        = 0.0f;
static ParityType       gCameraParity       = ParityNormal;

//  When simulating a manifold, keep a list of all group elements
//  close to the origin.  In the spherical case, keep track of *all*
//  group elements, because there are only finitely many. 
static int              gNumGroupElements       = 0;
static D3DXMATRIX       *gGroupElements         = NULL;
static ParityType       *gGroupElementParities  = NULL;

//  This sample code uses only one scenery object, namely a
//  spinning earth.  A more interesting program -- say a game
//  in a hypersphere -- would use a variety of scenery objects.
//  CurvedSpaces-Simple-D3D.cpp uses a SceneryObject class that may
//  be useful in more substantial programs.
static CSpinningEarth   *gSpinningEarth = NULL;

//  Different kinds of scenery will require different textures.
//  Be sure to keep this typedef synchronized with TextureType above.
static char *gTextureFiles[NumTextures] =
{
    "Textures\\EarthNorthern.bmp",
    "Textures\\EarthSouthern.bmp",
    "Textures\\Beam.bmp",
    "Textures\\Spaceship.bmp"
};
static LPDIRECTDRAWSURFACE7 gTextures[NumTextures] = {NULL, NULL, NULL, NULL};

//  For a realistic simulation, the program's internal parameters
//  must match the relative proportions of the user's physical display device,
//  the distance from the display to the user's face, and the separation
//  between the user's eyes.
//  Initialize the parameters for a 90 degree field of view.
//  In monoscopic 3D this gives a nice wide-angle view.
//  In stereoscopic 3D a 90 degree field of view will be practical
//  only with a very large display -- for example in a lecture hall --
//  relative to which the separation between the user's eyes is small.
//  Accurate stereoscopic 3D is more difficult on an ordinary monitor,
//  but is workable in the flat space case if you set gWindowWidth and
//  gWindowDist correctly, but underreport the separation between your eyes.
//  When the user activates Stereo at runtime, he or she is given
//  an opportunity to override the defaults given here.
static float    gWindowWidth    = 200.0f,
                gWindowDist     = 100.0f,
                gEyeSpacing     =   5.0f,
                gIntrinsicUnit  = 400.0f;   //  typically 2*gWindowWidth

//  The nine possible projection matrices are indexed
//  as gProjectionMatrix[EyeType][RangeType].
static D3DXMATRIX   gProjectionMatrix[3][3];

//  The program uses a single window.
static HWND gWindow = NULL;

//  The program maintains a D3DXContext, which is replaced
//  whenever the user changes the bit depth.
static ID3DXContext *gD3DXContext = NULL;

//  A Dirichlet domain provides a framework in which to understand
//  the shape of a manifold.
static CDirichletDomain     *gDirichletDomain = NULL;

//  Let the user change the fog color in mono mode
//  (but not in stereo, which requires a black background).
static COLORREF             gFogColorMono = FOG_COLOR_MONO;

//  The group elements defining a manifold are read from a file.
static char         gFileName[MAX_PATH + 1] = "";
static OPENFILENAME gFileInfo = {sizeof(OPENFILENAME), NULL, NULL,
    "Generator files (*.gen)\0*.gen\0Matrix files (*.mat)\0*.mat\0",
    NULL, 0, 0, NULL, MAX_PATH, NULL, 0, NULL, NULL, 0, 0, 0,
    "gen", 0L, NULL, NULL};

//  We'll want to return to our home directory when reading texture files.
static char         gHomeDirectory[256];

//  Viewing options.
static BOOL gViewEarth  = TRUE,
            gViewBeams  = TRUE;


static void             InitProjectionMatrices(void);
static void             BuildWindow(HINSTANCE hInstance);
static void             RestartGraphics(void);
static HRESULT          InitGraphics(void);
static void             LoadTexture(LPDIRECT3DDEVICE7 aD3DDevice, LPDIRECTDRAWSURFACE7 *aTexturePointer, char *aTextureFileName);
static void             ShutDownGraphics(void);
static LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
static void             MenuCommand(WORD aCommand, HWND aParentWindow);
static void             ChooseFogColor(void);
static BOOL FAR PASCAL  TrivialDialogProc(HWND hDlg, UINT msg, WPARAM wparam, LPARAM lParam);
static BOOL FAR PASCAL  EnvironmentDialogProc(HWND hDlg, UINT msg, WPARAM wparam, LPARAM lParam);
static void             ReadFile(char *aFileName);
static void             ReadGeneratorFile(char *aFileName);
static BOOL             SameMatrix(D3DXMATRIX *a, D3DXMATRIX *b);
static void             ReadMatrixFile(char *aFileName);
static int              CompareSpherical (const void *m1, const void *m2);
static int              CompareFlat      (const void *m1, const void *m2);
static int              CompareHyperbolic(const void *m1, const void *m2);
static void             Update(void);
static void             NormalizeCameraPosition(void);
static void             Render(void);
static void             DrawSpaceshipBackground(LPDIRECT3DDEVICE7 aD3DDevice);
static void             ProjectAndDraw(LPDIRECT3DDEVICE7 aD3DDevice, EyeType anEyeType);
static void             DrawContents(LPDIRECT3DDEVICE7 aD3DDevice, EyeType anEyeType, RangeType aRangeType);
static DWORD            ComputeFogFactor(float *u, float *v, int aRangeType, float *aDistance);
static double           ArcCosh(double x);
static void             HandleWindowedModeChanges(ID3DXContext *aD3DXContext);
static void             PrependTranslation(D3DXMATRIX *aMatrix, int anAxis, double aDistance);
static void             PrependRotation(D3DXMATRIX *aMatrix, float aYaw, float aPitch, float aRoll);
static void             InterpretError(HRESULT hr);


int PASCAL WinMain(
    HINSTANCE   hInstance,
    HINSTANCE   hPrevInstance, 
    LPSTR       lpszCmdLine,
    int         nCmdShow) 
{
    HRESULT hr;
    MSG     theMessage;

    //  Initialize the SpinningEarth.
    gSpinningEarth = new CSpinningEarth();
    if (gSpinningEarth == NULL)
        goto CleanUpAndExit;

    //  Note our home directory, for later use in finding textures.
    _getcwd(gHomeDirectory, 255);

    //  Initialize the window.
    BuildWindow(hInstance);

    //  Initialize the Direct3DX graphics.
    if (FAILED(hr = InitGraphics()))
    {
        InterpretError(hr);
        goto CleanUpAndExit;
    }

    //  Process messages until WM_QUIT is received.
    //  Render the scene during idle time.
    while (TRUE)
    {
        if (PeekMessage(&theMessage, NULL, 0, 0, PM_REMOVE) != 0)
        {
            TranslateMessage(&theMessage);
            DispatchMessage(&theMessage);

            if (theMessage.message == WM_QUIT)
                break;
        }
        else
        {
            if (IsIconic(gWindow) == FALSE
             && gNumGroupElements > 0)
            {
                Update();
                Render();
			}
            else
                Sleep(100);
       }
    }

CleanUpAndExit:

    //  Shutdown Direct3D.
    ShutDownGraphics();

    //  Release the window.
    if (gWindow != NULL)
        DestroyWindow(gWindow);

    //  Delete the SpinningEarth.
    SAFE_DELETE(gSpinningEarth)

    //  Delete the GroupElements.
    SAFE_DELETE_(gGroupElements);
    SAFE_DELETE_(gGroupElementParities);

    //  Delete the Dirichlet domain.
    SAFE_DELETE(gDirichletDomain);

    return 0;
}


static void InitProjectionMatrices(void)
{
    RECT    theClientRect;
    float   theWindowHeight;
    int     i,
            j,
            k,
            l;

    //  Deduce theWindowHeight from the aspect ratio
    //  of gWindow's client area.
    if (gWindow != NULL)
    {
        GetClientRect(gWindow, &theClientRect);
        theWindowHeight = (gWindowWidth * theClientRect.bottom) / theClientRect.right;
    }
    else
        theWindowHeight = gWindowWidth;

    //  Start with standard banana matrices.  By default the program
    //  begins with aWindowWidth == 2*aWindowDistance, giving a nice
    //  wide 90 degree field of view.  It's easy to verify that the
    //  standard banana matrix takes the corners of the viewing banana
    //  to the corners of the clipping box in projection space.
    //  Please see the accompanying article "Real-Time Rendering
    //  in Curved Spaces" for details.
    for (i = 0; i < 3; i++)
        for (j = 0; j < 3; j++)
        {
            D3DXMatrixIdentity(&gProjectionMatrix[i][j]);

            gProjectionMatrix[i][j].m[0][0] = 2 * gWindowDist / gWindowWidth;
            gProjectionMatrix[i][j].m[1][1] = 2 * gWindowDist / theWindowHeight;

            gProjectionMatrix[i][j].m[2][2] =  0.5f;
            gProjectionMatrix[i][j].m[2][3] =  1.0f;
            gProjectionMatrix[i][j].m[3][2] = -0.5f * CLIPPING_DISTANCE;
            gProjectionMatrix[i][j].m[3][3] =  0.0f;
        }

    //  Modify the BananaFront and BananaBack matrices to map into
    //  the front and back halves of the clipping box, respectively.
    //  Again, please see the main article for a complete explanation.
    for (i = 0; i < 3; i++)
    {
        gProjectionMatrix[i][BananaFront].m[2][2] =  0.25f;
        gProjectionMatrix[i][BananaFront].m[3][2] = -0.25f * CLIPPING_DISTANCE;

        gProjectionMatrix[i][BananaBack ].m[2][2] =  0.75f;
        gProjectionMatrix[i][BananaBack ].m[3][2] = -0.25f * CLIPPING_DISTANCE;
    }

    //  Offset the EyeLeft and EyeRight projections, to account for
    //  the fact that the left eye sees the window a little to its right,
    //  and the right eye sees it a little to its left.  The guiding
    //  principle is that a person using the program should receive
    //  the same light from the same directions as would a person
    //  in the actual spherical, flat, or hyperbolic space itself.
    for (j = 0; j < 3; j++)
    {
        gProjectionMatrix[EyeLeft ][j].m[2][0] = - gEyeSpacing / gWindowWidth;
        gProjectionMatrix[EyeRight][j].m[2][0] = + gEyeSpacing / gWindowWidth;
    }

    //  Multiply the BananaBack matrices by -Id to invert the scenery
    //  and eliminate the need to invert the camera at render time.
    for (i = 0; i < 3; i++)
        for (k = 0; k < 4; k++)
            for (l = 0; l < 4; l++)
                gProjectionMatrix[i][BananaBack].m[k][l] *= -1;

    //  Strictly speaking, the projection matrices are now complete.
    //  However, to save ourselves the trouble of offsetting the camera
    //  for the EyeLeft and EyeRight stereo views at render time, we can
    //  instead offset the scenery, and pre-multiply the offset into the
    //  projection matrices.
    for (j = 0; j < 3; j++)
    {
        PrependTranslation( &gProjectionMatrix[EyeLeft][j],
                            0,
                            + (gEyeSpacing / gIntrinsicUnit) / 2);

        PrependTranslation( &gProjectionMatrix[EyeRight][j],
                            0,
                            - (gEyeSpacing / gIntrinsicUnit) / 2);
    }
}


static void BuildWindow(
    HINSTANCE   hInstance)
{
    int theClientWidth,
        theClientHeight;

    //  Register the window class.
    WNDCLASS wndClass =
    {
        CS_HREDRAW | CS_VREDRAW,
        WndProc,
        0,
        0,
        hInstance,
        LoadIcon(hInstance, "CurvedSpacesCompleteIcon"),
        LoadCursor(hInstance, "CurvedSpacesCursor"),
        (HBRUSH)GetStockObject(BLACK_BRUSH),
        TEXT("CurvedSpacesMenu"),
        TEXT("Curved Spaces Window Class")
    };
    RegisterClass(&wndClass);

    //  Figure out the largest size square window
    //  that will fit comfortably on the screen.
    theClientWidth  =     GetSystemMetrics(SM_CXSCREEN)
                    - 2 * GetSystemMetrics(SM_CXSIZEFRAME)
                    - 2 * WINDOW_MARGIN;
    theClientHeight =     GetSystemMetrics(SM_CYSCREEN)
                    -     GetSystemMetrics(SM_CYCAPTION)
                    -     GetSystemMetrics(SM_CYMENU)
                    - 2 * GetSystemMetrics(SM_CYSIZEFRAME)
                    - 2 * WINDOW_MARGIN;
    if (theClientWidth > theClientHeight)
        theClientWidth = theClientHeight;
    if (theClientHeight > theClientWidth)
        theClientHeight = theClientWidth;

    //  Create the window.
    gWindow = CreateWindow (
        TEXT("Curved Spaces Window Class"),
        TEXT("Curved Spaces (Direct3D)"),
        WS_OVERLAPPEDWINDOW | WS_VISIBLE,
        WINDOW_MARGIN, WINDOW_MARGIN,
        theClientWidth  + 2 * GetSystemMetrics(SM_CXSIZEFRAME),
        theClientHeight + GetSystemMetrics(SM_CYCAPTION)
                        + GetSystemMetrics(SM_CYMENU)
                        + 2 * GetSystemMetrics(SM_CYSIZEFRAME),
        NULL, NULL, hInstance, NULL);
}



static void RestartGraphics(void)
{
    if (IsIconic(gWindow))
        return;

    ShutDownGraphics();

    if (FAILED(InitGraphics()))
    {
        MessageBox(NULL, "Can't reinitialize graphics.  Bye.", "We're cooked.", MB_OK);
        PostQuitMessage(0);
    }
}


static HRESULT InitGraphics(void)
{
    HRESULT             hr;
    LPDIRECT3DDEVICE7   theD3DDevice;
    int                 i;

    static DWORD        theHardwareSupportLevel[3] =
    {
        D3DX_HWLEVEL_TL,        //  hardware T&L and rasterization
        D3DX_HWLEVEL_RASTER,    //  hardware rasterization
        D3DX_HWLEVEL_2D         //  software rendering
    };

    //  Initialize the Direct3DX graphics.
    if (FAILED(hr = D3DXInitialize()))
        return hr;

    //  Create a Direct3DX Context.
    //
    //  Technical Note:  In theory we should be able to pass
    //  D3DX_DEFAULT for the deviceIndex and let D3DXCreateContext()
    //  pick the best available level of hardware support.
    //  In practice Direct3D sometimes shoots for a level of support
    //  that isn't available at the current bit depth, and then fails.
    //  For example, on my own graphics card it succeeds with
    //  16-bit color and fails with 32-bit color.
    //  To be safe, let's step through the support levels manually.
    i = 0;
    do
    {
        hr = D3DXCreateContext( theHardwareSupportLevel[i],
                                0,
                                gWindow,
                                D3DX_DEFAULT,
                                D3DX_DEFAULT,
                                &gD3DXContext);

    } while (FAILED(hr) && ++i < 3);

    if (FAILED(hr))
        return hr;

    //  Load any desired textures.
    //  If we can't load the textures, push on without them.
    theD3DDevice = gD3DXContext->GetD3DDevice();
    _chdir(gHomeDirectory);
    if (theD3DDevice != NULL)
    {
        for (i = 0; i < NumTextures; i++)
            LoadTexture(theD3DDevice, &gTextures[i], gTextureFiles[i]);
    }
    else
    {
        MessageBox(NULL, "Couldn't get a device context to load texture(s).", "Trouble", MB_OK);
        for (i = 0; i < NumTextures; i++)
            gTextures[i] = NULL;
    }
    SAFE_RELEASE(theD3DDevice);

    return S_OK;
}

static void LoadTexture(
    LPDIRECT3DDEVICE7       aD3DDevice,
    LPDIRECTDRAWSURFACE7    *aTexturePointer,
    char                    *aTextureFileName)
{
    HRESULT             hr;
    char                theBuffer[512];
    D3DX_SURFACEFORMAT  theSurfaceFormat;

    theSurfaceFormat = D3DX_SF_UNKNOWN;

    hr = D3DXCreateTextureFromFile(
        aD3DDevice,
        NULL,               // dwFlags
        NULL,               // auto-width
        NULL,               // auto-height
        &theSurfaceFormat,  // desired format
        NULL,               // pointer to Palette
        aTexturePointer,    // returned pointer to texture
        NULL,               // returned number of mipmaps
        aTextureFileName,   // file name for texture
        D3DX_FT_DEFAULT);   // default scaling
    if (FAILED(hr))
    {
        wsprintf(theBuffer, "Couldn't load texture from file %s.  Most likely the file is missing or your graphics card lacks sufficient memory.", aTextureFileName);
        MessageBox(NULL, theBuffer, "Texture Failure", MB_OK);
        *aTexturePointer = NULL;
    }
}


static void ShutDownGraphics(void)
{
    int i;

    //  Release any textures.
    for (i = 0; i < NumTextures; i++)
        SAFE_RELEASE(gTextures[i])

    //  Release the D3DX Context.
    SAFE_RELEASE(gD3DXContext)

    //  Shut down D3DX.
    D3DXUninitialize();
}


//  Callback function to process window messages.
static LRESULT CALLBACK WndProc(
    HWND    hWnd,
    UINT    uMsg,
    WPARAM  wParam,
    LPARAM  lParam)
{
    RECT            theRect;
    POINT           theNewCursorLocation;
    float           dx,
                    dy;
    static POINT    theWindowCenter = {0,0};
    static BOOL     theCursorIsHidden = FALSE;

    switch (uMsg)
    {
        case WM_CLOSE:
            PostQuitMessage(0);
            return 0;

        case WM_DESTROY:
            gWindow = NULL;
            return 0;

        case WM_KEYDOWN:
            switch (wParam) //  virtual key code
            {
                case VK_UP:
                    gCameraSpeed += CAMERA_SPEED_INCREMENT;
                    break;

                case VK_DOWN:
                    gCameraSpeed -= CAMERA_SPEED_INCREMENT;
                    break;

                case ' ':
                    gCameraSpeed = 0;
                    break;

                case VK_ESCAPE:
                    //  The user wants to stop flying.
                    //  ReleaseCapture() generates a WM_CAPTURECHANGED message
                    //  which will be handled below.
                    ReleaseCapture();
                    break;
            }
            return 0;

        case WM_COMMAND:
            MenuCommand(LOWORD(wParam), hWnd);
            return 0;

        case WM_LBUTTONDOWN:
            //  The user is clicking into the scene to steer.
            //  Capture the mouse and hide the cursor.
            SetCapture(hWnd);
            SetCursor(NULL);
            GetWindowRect(hWnd, &theRect);
            theWindowCenter.x = (theRect.right  + theRect.left)/2;
            theWindowCenter.y = (theRect.bottom + theRect.top )/2;
            SetCursorPos(theWindowCenter.x, theWindowCenter.y);
            theCursorIsHidden = TRUE;
            return 0;

        case WM_RBUTTONDOWN:
            //  The user wants to stop flying.
            //  ReleaseCapture() generates a WM_CAPTURECHANGED message
            //  which will be handled below.
            ReleaseCapture();
            return 0;

        case WM_CAPTURECHANGED:
            //  The WM_CAPTURECHANGED message typically results
            //  from our own call to ReleaseCapture(),
            //  but it could arise in other ways, for example,
            //  if some other program grabs the mouse.
            theCursorIsHidden = FALSE;
            return 0;

        case WM_MOUSEMOVE:
            //  When the cursor is hidden, mouse movements steer the camera.
            if (theCursorIsHidden)
            {
                GetCursorPos(&theNewCursorLocation);

                if (theNewCursorLocation.x != theWindowCenter.x
                 || theNewCursorLocation.y != theWindowCenter.y)
                {
                    GetWindowRect(hWnd, &theRect);

                    dx = (float)(theNewCursorLocation.x - theWindowCenter.x)
                       / (float)(theRect.right  - theRect.left);
                    dy = (float)(theNewCursorLocation.y - theWindowCenter.y)
                       / (float)(theRect.bottom - theRect.top );

                    SetCursorPos(theWindowCenter.x, theWindowCenter.y);

                    //  Allow full six-degrees-of-freedom navigation.
                    if (wParam & MK_SHIFT)
                    {
                        //  Translate
                        if (wParam & MK_CONTROL)
                            PrependTranslation(&gCameraPosition, 2, -dy);
                        else
                        {
                            PrependTranslation(&gCameraPosition, 0,  dx);
                            PrependTranslation(&gCameraPosition, 1, -dy);
                        }
                    }
                    else
                    {
                        //  Rotate
                        if (wParam & MK_CONTROL)   
                            PrependRotation(&gCameraPosition, 0, 0, dx);                         
                        else
                            PrependRotation(&gCameraPosition, dx, dy, 0);
                    }
                }
            }
            return 0;

        case WM_MOUSEWHEEL:
            gCameraSpeed += CAMERA_SPEED_INCREMENT * ((short)HIWORD(wParam))/120.0f;
            return 0;

        case WM_MBUTTONDOWN:    //  perhaps user has pressed mouse wheel
            gCameraSpeed = 0;
            return 0;

        case WM_SIZE:
            InitProjectionMatrices();
            if (gD3DXContext != NULL)
                gD3DXContext->Resize(LOWORD(lParam), HIWORD(lParam));
            return 0;

        case WM_GETMINMAXINFO:
            // Prevent the client area from going smaller
            //  than MIN_CLIENT_SIZE.
            ((MINMAXINFO*)lParam)->ptMinTrackSize.x =
                MIN_CLIENT_SIZE
                + 2 * GetSystemMetrics(SM_CXSIZEFRAME);
            ((MINMAXINFO*)lParam)->ptMinTrackSize.y =
                MIN_CLIENT_SIZE
                + GetSystemMetrics(SM_CYCAPTION)
                + GetSystemMetrics(SM_CYMENU)
                + 2 * GetSystemMetrics(SM_CYSIZEFRAME);
            return 0;

        case WM_INITMENUPOPUP:

            CheckMenuItem(  (HMENU)wParam, ID_VIEW_EARTH,
                gViewEarth            ? MF_CHECKED : MF_UNCHECKED);
            CheckMenuItem(  (HMENU)wParam, ID_VIEW_BEAMS,
                gViewBeams            ? MF_CHECKED : MF_UNCHECKED);
            CheckMenuItem(  (HMENU)wParam, ID_OPTIONS_FOG,
                gFog                  ? MF_CHECKED : MF_UNCHECKED);
            CheckMenuItem(  (HMENU)wParam, ID_OPTIONS_STEREO,
                gStereo               ? MF_CHECKED : MF_UNCHECKED);
            CheckMenuItem(  (HMENU)wParam, ID_OPTIONS_BACKSIDE,
                gRenderBackHemisphere ? MF_CHECKED : MF_UNCHECKED);
            CheckMenuItem(  (HMENU)wParam, ID_OPTIONS_FRAMERATE,
                gShowFrameRate        ? MF_CHECKED : MF_UNCHECKED);

           return 0;

        default:
            return DefWindowProc(hWnd, uMsg, wParam, lParam);
    }
}


static void MenuCommand(
    WORD    aCommand,
    HWND    aParentWindow)
{
    switch (aCommand)
    {
        //  file menu

        case ID_FILE_OPEN:
            gFileInfo.hwndOwner         = gWindow;
            gFileInfo.lpstrFile         = gFileName;
            gFileInfo.Flags             = 0;
            if (GetOpenFileName(&gFileInfo))
                ReadFile(gFileName);
            break;

        case ID_FILE_EXIT:
            PostQuitMessage(0);
            break;

        //  view menu

        case ID_VIEW_EARTH:
            gViewEarth = ! gViewEarth;
            break;

        case ID_VIEW_BEAMS:
            gViewBeams = ! gViewBeams;
            break;

        //  options menu

        case ID_OPTIONS_FOG:
            gFog = ! gFog;
            break;

        case ID_OPTIONS_FOGCOLOR:
            ChooseFogColor();
            break;

        case ID_OPTIONS_STEREO:
            gStereo = ! gStereo;
            break;

        case ID_OPTIONS_PERSPECTIVE:
            DialogBox(  (HINSTANCE)GetWindowLong(aParentWindow, GWL_HINSTANCE),
                        TEXT("EnvironmentDialog"),
                        aParentWindow, 
                        (DLGPROC)EnvironmentDialogProc);
            break;

        case ID_OPTIONS_BACKSIDE:
            gRenderBackHemisphere = ! gRenderBackHemisphere;
            break;

        case ID_OPTIONS_FRAMERATE:
            gShowFrameRate = ! gShowFrameRate;
            SetWindowText(gWindow, gShowFrameRate ? "Counting..." : "Curved Spaces (Direct 3D)");
            break;

        //  help menu

        case ID_HELP_INSTRUCTIONS:
            DialogBox(  (HINSTANCE)GetWindowLong(aParentWindow, GWL_HINSTANCE),
                        TEXT("INSTRUCTIONSDIALOG"),
                        aParentWindow, 
                        (DLGPROC)TrivialDialogProc);
            break;

        case ID_HELP_ABOUT:
            DialogBox(  (HINSTANCE)GetWindowLong(aParentWindow, GWL_HINSTANCE),
                        TEXT("ABOUTDIALOG"),
                        aParentWindow, 
                        (DLGPROC)TrivialDialogProc);
            break;
    }
}

static void ChooseFogColor(void)
{
    CHOOSECOLOR theChooseColorData;

    static COLORREF theCustomColors[16] = {
                0x00FFFFFF, 0x00FFFFFF, 0x00FFFFFF, 0x00FFFFFF,
                0x00FFFFFF, 0x00FFFFFF, 0x00FFFFFF, 0x00FFFFFF,
                0x00FFFFFF, 0x00FFFFFF, 0x00FFFFFF, 0x00FFFFFF,
                0x00FFFFFF, 0x00FFFFFF, 0x00FFFFFF, 0x00FFFFFF};

    theChooseColorData.lStructSize      = sizeof(CHOOSECOLOR);
    theChooseColorData.hwndOwner        = gWindow;
    theChooseColorData.hInstance        = NULL;
    theChooseColorData.rgbResult        = RGB2BRG(gFogColorMono);
    theChooseColorData.lpCustColors     = theCustomColors;
    theChooseColorData.Flags            = CC_RGBINIT;
    theChooseColorData.lCustData        = 0;
    theChooseColorData.lpfnHook         = NULL;
    theChooseColorData.lpTemplateName   = NULL;

    if (ChooseColor(&theChooseColorData))
        gFogColorMono = RGB2BRG(theChooseColorData.rgbResult);
}

static BOOL FAR PASCAL TrivialDialogProc(
    HWND    hDlg,
    UINT    msg,
    WPARAM  wparam,
    LPARAM  lParam)
{
    switch (msg)
    {
        case WM_INITDIALOG:
            return TRUE;

        case WM_COMMAND:
            switch (LOWORD(wparam))
            {
                case IDOK:
                    EndDialog(hDlg, TRUE);
                    return TRUE;

                default:
                    return FALSE;
            }

        default:
            return FALSE;
    }
}

static BOOL FAR PASCAL EnvironmentDialogProc(
    HWND    hDlg,
    UINT    msg,
    WPARAM  wparam,
    LPARAM  lParam)
{
    char    theBuffer[32];
    float   theNumber;

    switch (msg)
    {
        case WM_INITDIALOG:

            //  Copy our internal variables into the dialog's edit fields.

            sprintf(theBuffer, "%.0f", gWindowWidth);
            SetWindowText(GetDlgItem(hDlg, IDC_WINDOW_WIDTH   ), theBuffer);

            sprintf(theBuffer, "%.0f", gWindowDist);
            SetWindowText(GetDlgItem(hDlg, IDC_WINDOW_DISTANCE), theBuffer);

            sprintf(theBuffer, "%.1f", gEyeSpacing);
            SetWindowText(GetDlgItem(hDlg, IDC_EYE_SPACING    ), theBuffer);

            sprintf(theBuffer, "%.0f", gIntrinsicUnit);
            SetWindowText(GetDlgItem(hDlg, IDC_INTRINSIC_UNIT ), theBuffer);

            return TRUE;

        case WM_COMMAND:

            switch (LOWORD(wparam))
            {
                case IDOK:

                    //  Copy the dialog's edit fields back into our internal variables.

                    GetWindowText(GetDlgItem(hDlg, IDC_WINDOW_WIDTH   ), theBuffer, 32);
                    if (sscanf(theBuffer, "%f", &theNumber) == 1)
                        gWindowWidth = theNumber;

                    GetWindowText(GetDlgItem(hDlg, IDC_WINDOW_DISTANCE), theBuffer, 32);
                    if (sscanf(theBuffer, "%f", &theNumber) == 1)
                        gWindowDist = theNumber;

                    GetWindowText(GetDlgItem(hDlg, IDC_EYE_SPACING    ), theBuffer, 32);
                    if (sscanf(theBuffer, "%f", &theNumber) == 1)
                        gEyeSpacing = theNumber;

                    GetWindowText(GetDlgItem(hDlg, IDC_INTRINSIC_UNIT ), theBuffer, 32);
                    if (sscanf(theBuffer, "%f", &theNumber) == 1)
                        gIntrinsicUnit = theNumber;

                    //  Dismiss the dialog.
                    EndDialog(hDlg, TRUE);

                    //  Recompute the projection matrices using the new parameters.
                    InitProjectionMatrices();

                    return TRUE;

                case IDCANCEL:
                    EndDialog(hDlg, TRUE);
                    return TRUE;

                default:
                    return FALSE;
            }
            return TRUE;
    }

    return FALSE;
}


static void ReadFile(char *aFileName)
{
    char    *theFileExtension;

    theFileExtension = aFileName + strlen(aFileName) - 3;

    if (strcmp(theFileExtension, "gen") == 0)
        ReadGeneratorFile(aFileName);

    if (strcmp(theFileExtension, "mat") == 0)
        ReadMatrixFile(aFileName);
}


static void ReadGeneratorFile(char *aFileName)
{
    FILE        *theFile;
    double      theCutoffRadius,
                theTranslationDistance;
    float       theDummy;
    int         theCount,
                theResult,
                theNumGenerators,
                theNumMatrices,
                i,
                j,
                k;
    D3DXMATRIX  *theGenerators,
                *theMatrices,
                theCandidate,
                theMatrix;
    MatrixNode  *theFirstNode,
                *theSecondNode,
                *theNode,
                *theExistingNode,
                *theLastNode;
    ParityType  *theParities;
    SpaceType   theSpaceType,
                theNewSpaceType;
    BOOL        theCandidateIsOnTheList;
    char        *theWindowTitle,
                *theChar;

    theFile         = NULL;
    theGenerators   = NULL;
    theMatrices     = NULL;
    theFirstNode    = NULL;
    theParities     = NULL;

    SetCursor(LoadCursor(NULL, IDC_WAIT));

    try
    {
        //  Open the file.
        theFile = fopen(aFileName, "r");
        if (theFile == NULL)
            throw new CException("Couldn't open file.");

        //  Read the cutoff radius.
        if (1 != fscanf(theFile, "%lf", &theCutoffRadius))
            throw new CException("Couldn't read cutoff radius.");

        //  Count the total number of matrix entries in the file.
        theCount = 0;
        do
        {
            theResult = fscanf(theFile, "%f", &theDummy);
            if (theResult == 1)
                theCount++;
            else if (theResult != EOF)
                throw new CException("File contains non-numeric entries.");
        } while (theResult == 1);

        //  Reset the file pointer to the first matrix entry.
        rewind(theFile);
        fscanf(theFile, "%lf", &theCutoffRadius);

        //  The number of entries should be a multiple of 16.
        if (theCount%16 != 0)
            throw new CException("File format:  cutoff radius followed by 4x4 generator matrices.");

        //  How many matrices does the file contain?
        theNumGenerators = theCount/16;

        //  Allocate space for the generators and their inverses.
        theGenerators = new D3DXMATRIX[2*theNumGenerators];
        if (theGenerators == NULL)
            throw new CException("Memory allocation failed.");

        //  Read the generators.
        for (i = 0; i < theNumGenerators; i++)
        {
            for (j = 0; j < 4; j++)
                for (k = 0; k < 4; k++)
                    if (1 != fscanf(theFile, "%f", &theGenerators[2*i].m[j][k]))
                        throw new CException("Error reading generators.");

            if (theGenerators[2*i].m[3][3] <  1.0f)
                theNewSpaceType = SpaceSpherical;
            if (theGenerators[2*i].m[3][3] == 1.0f)
                theNewSpaceType = SpaceFlat;
            if (theGenerators[2*i].m[3][3] >  1.0f)
                theNewSpaceType = SpaceHyperbolic;

            if (i == 0)
                theSpaceType = theNewSpaceType;
            else
            {
                if (theSpaceType != theNewSpaceType)
                    throw new CException("Matrices belong to inconsistent geometries. (Or perhaps the identity was unnecessarily included in the file.)");
            }

            //  Compute the generator's inverse.
            switch (theSpaceType)
            {
                case SpaceSpherical:
                    D3DXMatrixTranspose(&theGenerators[2*i+1], &theGenerators[2*i]);
                    break;
                case SpaceFlat:
                    D3DXMatrixInverse(&theGenerators[2*i+1], NULL, &theGenerators[2*i]);
                    break;
                case SpaceHyperbolic:
                    D3DXMatrixTranspose(&theGenerators[2*i+1], &theGenerators[2*i]);
                    for (j = 0; j < 4; j++)
                        for (k = 0; k < 4; k++)
                            if ((j == 3) != (k == 3))
                                theGenerators[2*i+1].m[j][k] = - theGenerators[2*i+1].m[j][k];
                    break;
            }
        }

        //  Close the file.
        fclose(theFile);
        theFile = NULL;

        //  Find all group elements out to theCutoffRadius
        //  and put them on a NULL-terminated singly-linked list.

        theFirstNode = new MatrixNode;
        if (theFirstNode == NULL)
            throw new CException("Memory allocation failed.");
        D3DXMatrixIdentity(&theFirstNode->d3dm);
        theFirstNode->next = NULL;
        theLastNode = theFirstNode;
        theNumMatrices = 1;

        for (   theNode = theFirstNode;
                theNode != NULL;
                theNode = theNode->next)
        {
            for (i = 0; i < 2*theNumGenerators; i++)
            {
                //  Multiply the given matrix by the given generator
                //  to obtain theCandidate.
                D3DXMatrixMultiply( &theCandidate,
                                    &theGenerators[i],
                                    &theNode->d3dm);

                //  Is theCandidate's translation distance too great?
                switch (theSpaceType)
                {
                    case SpaceSpherical:
                        theTranslationDistance = acos(theCandidate.m[3][3]);
                        break;

                    case SpaceFlat:
                        theTranslationDistance = sqrt(
                                theCandidate.m[3][0]*theCandidate.m[3][0]
                              + theCandidate.m[3][1]*theCandidate.m[3][1]
                              + theCandidate.m[3][2]*theCandidate.m[3][2]);
                        break;

                    case SpaceHyperbolic:
                        theTranslationDistance = ArcCosh(theCandidate.m[3][3]);
                        break;
                }
                if (theTranslationDistance > theCutoffRadius)
                    continue;

                //  Is theCandidate already on the list?
                theCandidateIsOnTheList = FALSE;
                for (   theExistingNode = theFirstNode;
                        theExistingNode != NULL;
                        theExistingNode = theExistingNode->next)
                    if (SameMatrix(&theCandidate, &theExistingNode->d3dm) == TRUE)
                    {
                        theCandidateIsOnTheList = TRUE;
                        break;
                    }

                //  If the candidate isn't already on the list,
                //  add it at the end.
                if (theCandidateIsOnTheList == FALSE)
                {
                    theLastNode->next = new MatrixNode;
                    if (theLastNode->next == NULL)
                        throw new CException("Memory allocation failed.");
                    theLastNode->next->d3dm = theCandidate;
                    theLastNode->next->next = NULL;
                    theLastNode = theLastNode->next;
                    theNumMatrices++;
                }
            }
        }

        //  We're done with theGenerators.
        SAFE_DELETE_(theGenerators);

        //  Copy the matrices to an array.
        theMatrices = new D3DXMATRIX[theNumMatrices];
        if (theMatrices == NULL)
            throw new CException("Memory allocation failed.");
        i = 0;
        for (   theNode = theFirstNode;
                theNode != NULL;
                theNode = theNode->next)
        {
            theMatrices[i++] = theNode->d3dm;
        }
        if (i != theNumMatrices)
            throw new CException("Internal error involving theNumMatrices.");

        //  Free the linked list.
        while (theFirstNode != NULL)
        {
            theSecondNode = theFirstNode->next;
            SAFE_DELETE(theFirstNode);
            theFirstNode = theSecondNode;
        }

        //  Because Direct3D can't write into the red and blue-green
        //  channels separately, stereoscopic rendering relies on
        //  a kludge that works best when the groups elements
        //  are sorted near to far.
        switch (theSpaceType)
        {
            case SpaceSpherical:
                qsort(theMatrices, theNumMatrices, sizeof(D3DXMATRIX), &CompareSpherical);
                break;
            case SpaceFlat:
                qsort(theMatrices, theNumMatrices, sizeof(D3DXMATRIX), &CompareFlat);
                break;
            case SpaceHyperbolic:
                qsort(theMatrices, theNumMatrices, sizeof(D3DXMATRIX), &CompareHyperbolic);
                break;
        }

        //  Allocate space for the parities.
        theParities = new ParityType[theNumMatrices];
        if (theParities == NULL)
            throw new CException("Memory allocation failed.");

        //  Compute the parity of each group element.
        for (i = 0; i < theNumMatrices; i++)
            theParities[i] =
                (D3DXMatrixfDeterminant(&theMatrices[i]) > 0.0f ?
                    ParityNormal : ParityReversed);

        //  Mathematicians using this program may create
        //  their own spherical manifolds by combining
        //  a left-handed and a right-handed group.
        //  Let's do them a favor and check that the result
        //  contains no group element with fixed points
        //  except the identity.
        //  To test for fixed points it's enough to test whether
        //  1 is a root of the characteristic polynomial, which
        //  is equivalent to testing whether det(g - I) is zero.
        //
        //  Comment out these lines to allow for orbifolds.
#if 0
        if (theSpaceType == SpaceSpherical)
            for (i = 1; i < theNumMatrices; i++)
            {
                theMatrix = theMatrices[i];
                for (j = 0; j < 4; j++)
                    theMatrix.m[j][j] -= 1.0f;
                if (fabs(D3DXMatrixfDeterminant(&theMatrix)) < SINGULAR_EPSILON)
                    throw new CException("Group contains fixed points.");
            }
#endif

        //  We've now finished all the stuff that might fail,
        //  such as reading the file and allocating memory.
        //  So let's drop out of the try{} block.
    }
    catch (CException *anException)
    {
        MessageBox(NULL, anException->Message(), "File Error", MB_OK);
        SAFE_DELETE(anException)

        if (theFile != NULL)
            fclose(theFile);
        SAFE_DELETE_(theGenerators);
        SAFE_DELETE_(theMatrices);
        SAFE_DELETE_(theParities);
        while (theFirstNode != NULL)
        {
            theSecondNode = theFirstNode->next;
            SAFE_DELETE(theFirstNode);
            theFirstNode = theSecondNode;
        }

        return;
    }

    //  Install the new matrices.
    SAFE_DELETE_(gGroupElements);
    SAFE_DELETE_(gGroupElementParities);
    gGroupElements          = theMatrices;
    gGroupElementParities   = theParities;
    gNumGroupElements       = theNumMatrices;
    gSpaceType              = theSpaceType;

    //  Recompute the projection matrices for the new SpaceType.
    InitProjectionMatrices();

    //  Reset the camera.
    D3DXMatrixIdentity(&gCameraPosition);
    gCameraParity   = ParityNormal;
//  gCameraSpeed    = 0.0f;     //  uncomment to reset camera speed

    //  Compute the new Dirichlet domain.
    SAFE_DELETE(gDirichletDomain);
    gDirichletDomain = new CDirichletDomain(
        gNumGroupElements,
        gGroupElements,
        (SpaceTypeDD) gSpaceType);  //  needed for consistent beam widths

    //  Set the file name as the window title,
    //  after skipping past the directory path and
    //  temporarily suppressing the file name extension.
    theWindowTitle = aFileName;
    for (theChar = aFileName; *theChar != 0; theChar++)
        if (*theChar == '\\')
            theWindowTitle = theChar + 1;
    theWindowTitle[strlen(theWindowTitle) - 4] = 0;
    SetWindowText(gWindow, theWindowTitle);
    theWindowTitle[strlen(theWindowTitle)] = '.';
}


static BOOL SameMatrix(
    D3DXMATRIX  *a,
    D3DXMATRIX  *b)
{
    int i,
        j;

    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            if (fabs(a->m[i][j] - b->m[i][j]) > MATRIX_EPSILON)
                return FALSE;

    return TRUE;
}


static void ReadMatrixFile(char *aFileName)
{
    int         theOldNumGroupElements;
    D3DXMATRIX  *theOldGroupElements;
    ParityType  *theOldGroupElementParities;
    FILE        *theFile;
    float       theDummy;
    int         theCount,
                theResult,
                i,
                j,
                k;
    float       theHeight;
    char        *theWindowTitle,
                *theChar;

    //  Read the matrices from the given file.

    //  Keep the old matrices around until the new ones have been read.
    theOldNumGroupElements      = gNumGroupElements;
    theOldGroupElements         = gGroupElements;
    theOldGroupElementParities  = gGroupElementParities;

    gNumGroupElements       = 0;
    gGroupElements          = NULL;
    gGroupElementParities   = NULL;

    try
    {
        //  Open the file.
        theFile = fopen(aFileName, "r");
        if (theFile == NULL)
            throw new CException("Couldn't open file.");

        //  Count the total number of matrix entries in the file.
        theCount = 0;
        do
        {
            theResult = fscanf(theFile, "%f", &theDummy);
            if (theResult == 1)
                theCount++;
            else if (theResult != EOF)
                throw new CException("File contains non-numeric entries.");
        } while (theResult == 1);
        rewind(theFile);

        //  The number of entries should be a positive multiple of 16.
        if (theCount%16 != 0)
            throw new CException("File must contain 4x4 matrices.");
        if (theCount == 0)
            throw new CException("File is empty.");

        //  How many matrices does the file contain?
        gNumGroupElements = theCount/16;

        //  Allocate space for the new matrices.
        gGroupElements = new D3DXMATRIX[gNumGroupElements];
        if (gGroupElements == NULL)
            throw new CException("Memory allocation failed.");

        //  Read the new matrices.
        for (i = 0; i < gNumGroupElements; i++)
            for (j = 0; j < 4; j++)
                for (k = 0; k < 4; k++)
                    fscanf(theFile, "%f", &gGroupElements[i].m[j][k]);

        //  Close the file.
        fclose(theFile);

        //  Allocate space for the parities.
        //  Note:  We allocate the space now while we're still
        //  inside the try{} block, but compute the parities later,
        //  after sorting the group elements.
        gGroupElementParities = new ParityType[gNumGroupElements];
        if (gGroupElementParities == NULL)
            throw new CException("Memory allocation failed.");

        //  Delete the old group elements.
        SAFE_DELETE_(theOldGroupElements);
        SAFE_DELETE_(theOldGroupElementParities);
    }
    catch (CException *anException)
    {
        MessageBox(NULL, anException->Message(), "File Error", MB_OK);
        SAFE_DELETE(anException)

        if (theFile != NULL)
            fclose(theFile);

        SAFE_DELETE_(gGroupElements);
        SAFE_DELETE_(gGroupElementParities);

        gNumGroupElements       = theOldNumGroupElements;
        gGroupElements          = theOldGroupElements;
        gGroupElementParities   = theOldGroupElementParities;

        return;
    }

    //  Deduce the SpaceType from the matrices.
    //  If the first matrix is the identity, we may
    //  need to look at the second matrix to figure it out.

    if (gGroupElements[0].m[3][3] != 1.0f || gNumGroupElements == 1)
        theHeight = gGroupElements[0].m[3][3];
    else
        theHeight = gGroupElements[1].m[3][3];

    if      (theHeight <  1.0f)  gSpaceType = SpaceSpherical;
    else if (theHeight == 1.0f)  gSpaceType = SpaceFlat;
    else                         gSpaceType = SpaceHyperbolic;

    //  Stereoscopic rendering works better when the groups elements
    //  are sorted near to far.
    switch (gSpaceType)
    {
        case SpaceSpherical:
            qsort(gGroupElements, gNumGroupElements, sizeof(D3DXMATRIX), &CompareSpherical);
            break;
        case SpaceFlat:
            qsort(gGroupElements, gNumGroupElements, sizeof(D3DXMATRIX), &CompareFlat);
            break;
        case SpaceHyperbolic:
            qsort(gGroupElements, gNumGroupElements, sizeof(D3DXMATRIX), &CompareHyperbolic);
            break;
    }

    //  Compute the parity of each group element.
    for (i = 0; i < gNumGroupElements; i++)
        gGroupElementParities[i] =
            (D3DXMatrixfDeterminant(&gGroupElements[i]) > 0.0f ?
                ParityNormal : ParityReversed);

    //  Recompute the projection matrices for the new SpaceType.
    InitProjectionMatrices();

    //  Reset the camera position, but maintain the previous speed.
    D3DXMatrixIdentity(&gCameraPosition);
    gCameraParity = ParityNormal;
//  gCameraSpeed    = 0.0f;     //  uncomment to reset camera speed

    //  Compute the new Dirichlet domain.
    SAFE_DELETE(gDirichletDomain);
    gDirichletDomain = new CDirichletDomain(
        gNumGroupElements,
        gGroupElements,
        (SpaceTypeDD) gSpaceType);  //  needed for consistent beam widths

    //  Set the file name as the window title,
    //  after skipping past the directory path and
    //  temporarily suppressing the file name extension.
    theWindowTitle = aFileName;
    for (theChar = aFileName; *theChar != 0; theChar++)
        if (*theChar == '\\')
            theWindowTitle = theChar + 1;
    theWindowTitle[strlen(theWindowTitle) - 4] = 0;
    SetWindowText(gWindow, theWindowTitle);
    theWindowTitle[strlen(theWindowTitle)] = '.';
}

static int CompareSpherical(const void *m1, const void *m2)
{
    float   a,
            b;

    a = ((D3DXMATRIX *) m1)->m[3][3];
    b = ((D3DXMATRIX *) m2)->m[3][3];

    if (a > b) return -1;
    if (a < b) return +1;
               return  0;
}

static int CompareFlat(const void *m1, const void *m2)
{
    float   *r,
            *s,
            a,
            b;

    r = ((D3DXMATRIX *) m1)->m[3];
    s = ((D3DXMATRIX *) m2)->m[3];

    a = r[0]*r[0] + r[1]*r[1] + r[2]*r[2];
    b = s[0]*s[0] + s[1]*s[1] + s[2]*s[2];

    if (a < b) return -1;
    if (a > b) return +1;
               return  0;
}

static int CompareHyperbolic(const void *m1, const void *m2)
{
    float   a,
            b;

    a = ((D3DXMATRIX *) m1)->m[3][3];
    b = ((D3DXMATRIX *) m2)->m[3][3];

    if (a < b) return -1;
    if (a > b) return +1;
               return  0;
}


static void Update(void)
{
    //  Tell the camera and each scenery object how many
    //  seconds have elapsed since the last Update().

    static unsigned int theStartTime    = 0,
                        theNumFrames    = 0;
    static double       theFrameRate    = 30.0;
    static BOOL         theFirstSecond  = TRUE;
    
    unsigned int    theCurrentTime,
                    theElapsedTime;
    double          theNewFrameRate;
    char            theWindowTitle[128];

    //  Get the time in milliseconds.
    theCurrentTime = GetTickCount();

    //  One-time initialization.
    if (theStartTime == 0)
    {
        theStartTime = theCurrentTime;
    }
    else
    {
        //  Count the number of frames drawn in approximately one second.
        theNumFrames++;
        theElapsedTime = theCurrentTime - theStartTime;
        if (theFirstSecond == TRUE && theElapsedTime > 0)
        {
            //  During the first second of operation,
            //  revise theFrameRate after each clock tick.
            //  At first it won't be super accurate, but at least
            //  it'll be better than our arbitrary initial guess.
            theFrameRate = theNumFrames / (0.001 * theElapsedTime);
        }
        if (theElapsedTime > 1000)
        {
            theFirstSecond = FALSE;

            theNewFrameRate = theNumFrames / (0.001 * theElapsedTime);
            theNumFrames = 0;
            theStartTime = theCurrentTime;
        
            //  If the user holds down a menu, our program gets suspended
            //  but the clock keeps running.  As a result, the nominal
            //  frame rate is too low and the scenery moves too fast
            //  in compensation.  To avoid this problem, don't let
            //  the nominal frame rate decrease too fast.
            if (theNewFrameRate < 0.8 * theFrameRate)
                theNewFrameRate = 0.8 * theFrameRate;
        
            theFrameRate = theNewFrameRate;

            if (gShowFrameRate)
            {
                wsprintf(theWindowTitle, "%3d frames per second", (int)theFrameRate);
                SetWindowText(gWindow, theWindowTitle);
            }
        }
    }

    //  Update the camera position.
    PrependTranslation(&gCameraPosition, 2, gCameraSpeed * (1.0 / theFrameRate));

    //  If a Dirichlet domain is present, keep the camera within
    //  the central cell.  (Note:  This isn't necessary in the
    //  spherical case -- where the whole space is finite -- but
    //  does no harm there.)
    if (gDirichletDomain != NULL)
    {
        gDirichletDomain->CenterCamera(&gCameraPosition);
        gCameraParity = 
            (D3DXMatrixfDeterminant(&gCameraPosition) > 0.0f ?
                ParityNormal : ParityReversed);
    }

    //  Normalize gCameraPosition to keep it legal for the current SpaceType.
    NormalizeCameraPosition();

    //  Update the SpinningEarth.
    gSpinningEarth->Update(1.0 / theFrameRate);
}


static void NormalizeCameraPosition(void)
{
    //  Numerical errors can accumulate and force the camera position
    //  matrix "out of round", in the sense that its rows are no
    //  longer orthonormal.  This effect is small in spherical and
    //  flat spaces, but can be significant in hyperbolic spaces,
    //  especially if the camera travels far from the origin.

    //  The "Gram-Schmidt process" consists of rescaling each
    //  row to restore unit length, and subtracting small multiples
    //  of one row from another to restore orthogonality.
    //  Here we carry out a first-order approximation to the
    //  Gram-Schmidt process.  That is, we normalize each row
    //  to unit length, but then assume that the subsequent
    //  orthogonalization step doesn't spoil the unit length.
    //  This assumption will be well satisfied because small
    //  first order changes orthogonal to a given vector affect
    //  its length only to second order.

    //  If your primary goal is to understand the techniques
    //  of curved space rendering, you may safely skip over
    //  the details of this orthonormalization. 

    D3DXMATRIX  m1,
                m2,
                theProduct;
    float       theFactor;
    int         i,
                j,
                k;

    //  Concoct a matrix giving the inner products
    //  of gCameraPosition's rows with one another.

    m1 = gCameraPosition;
    D3DXMatrixTranspose(&m2, &m1);

    switch (gSpaceType)
    {
        case SpaceSpherical:
            //  no change required
            break;

        case SpaceFlat:

            //  zero out the mixed terms in both m1 and m2

            m1.m[3][0] = 0.0f;
            m1.m[3][1] = 0.0f;
            m1.m[3][2] = 0.0f;

            m2.m[0][3] = 0.0f;
            m2.m[1][3] = 0.0f;
            m2.m[2][3] = 0.0f;

            break;

        case SpaceHyperbolic:

            //  negate the mixed terms in m2 only

            m2.m[0][3] = - m2.m[0][3];
            m2.m[1][3] = - m2.m[1][3];
            m2.m[2][3] = - m2.m[2][3];

            m2.m[3][0] = - m2.m[3][0];
            m2.m[3][1] = - m2.m[3][1];
            m2.m[3][2] = - m2.m[3][2];

            break;
    }

    D3DXMatrixMultiply(&theProduct, &m1, &m2);

    //  Normalize the length of each row of gCameraPosition.
    for (i = 0; i < 4; i++)
    {
        if (theProduct.m[i][i] <= 0.0f)
            exit(1);

        theFactor = (float) (1.0 / sqrt(theProduct.m[i][i]));

        for (j = 0; j < 4; j++)
            gCameraPosition.m[i][j] *= theFactor;
    }

    //  Make the rows of gCameraPosition mutually orthogonal.
    for (i = 4; --i >= 0; )         // iterate backwards
        for (j = i; --j >= 0; )     // iterate backwards
        {
            theFactor = theProduct.m[j][i];

            for (k = 0; k < 4; k++)
                gCameraPosition.m[j][k] -=
                    theFactor * gCameraPosition.m[i][k];
                    
        }
}


static void Render(void)
{
    LPDIRECT3DDEVICE7   theD3DDevice;
    D3DXMATRIX          theViewMatrix;
    LPDIRECTDRAW7       theDD = NULL;

    if (gWindow == NULL
     || IsIconic(gWindow))
        return;

    try
    {
        theD3DDevice = gD3DXContext->GetD3DDevice();
        if (theD3DDevice == NULL)
            throw new CException("Couldn't get D3DDevice from D3DContext.");

        //  Using w buffering instead of z buffering would give incorrect
        //  results in the spherical case, so let's explicitly set
        //  D3DRENDERSTATE_ZENABLE to D3DZB_TRUE (as opposed to
        //  D3DZBUFFERTYPE's third possibility, which is D3DZB_USEW).
        D3DTRY(theD3DDevice->SetRenderState(D3DRENDERSTATE_ZENABLE, D3DZB_TRUE))

        D3DTRY(theD3DDevice->SetRenderState(D3DRENDERSTATE_LIGHTING, FALSE))

        D3DTRY(theD3DDevice->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTFG_LINEAR));
        D3DTRY(theD3DDevice->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTFN_LINEAR));

        D3DTRY(theD3DDevice->SetRenderState(D3DRENDERSTATE_FOGENABLE, gFog))
        D3DTRY(theD3DDevice->SetRenderState(D3DRENDERSTATE_FOGCOLOR,  gStereo ? FOG_COLOR_STEREO : gFogColorMono))

        D3DTRY(theD3DDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, gStereo));
        D3DTRY(theD3DDevice->SetRenderState(D3DRENDERSTATE_SRCBLEND,         D3DBLEND_ONE));
        D3DTRY(theD3DDevice->SetRenderState(D3DRENDERSTATE_DESTBLEND,        D3DBLEND_ONE));

        D3DTRY(theD3DDevice->BeginScene())

        //  Initialize the render target and z-buffer.
        if (gSpaceType == SpaceSpherical
         && gStereo == FALSE)
        {
            //  In a spherical space we see ourselves in all directions,
            //  so the back of our own spaceship fills the sky.
            D3DTRY(theD3DDevice->SetRenderState(D3DRENDERSTATE_ZENABLE, D3DZB_FALSE))
            DrawSpaceshipBackground(theD3DDevice);
            D3DTRY(theD3DDevice->SetRenderState(D3DRENDERSTATE_ZENABLE, D3DZB_TRUE))
            D3DTRY(theD3DDevice->Clear(0, NULL, D3DCLEAR_ZBUFFER, 0x00000000, 1.0f, 0))
        }
        else
            //  Otherwise clear the render target to fog color.
            D3DTRY(theD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, gStereo ? FOG_COLOR_STEREO : gFogColorMono, 1.0f, 0))

        //  The view matrix (which maps world space into camera space) is the
        //  inverse of gCameraPosition (which maps camera space into world space).
        D3DXMatrixInverse(&theViewMatrix, NULL, &gCameraPosition);
        D3DTRY(theD3DDevice->SetTransform(
            D3DTRANSFORMSTATE_VIEW,
            &D3DMATRIX(theViewMatrix)))

        //  Standard rendering shows all scenery in flat and hyperbolic spaces.
        //  In a spherical space it shows only the front hemisphere, but if
        //  the scenery objects are distributed symmetrically (with each object
        //  having a twin at its antipodal point) nothing is lost by omitting
        //  the back hemisphere.
        //  Draw.
        if (gStereo == FALSE)
        {
            //  Draw a full color image for a single eye.
            ProjectAndDraw(theD3DDevice, EyeOnly);
        }
        else
        {
            //  Draw the left eye image.
            ProjectAndDraw(theD3DDevice, EyeLeft);

            //  Clear the z-buffer.
            D3DTRY(theD3DDevice->Clear(0, NULL, D3DCLEAR_ZBUFFER, 0x00000000, 1.0f, 0))

            //  Draw the right eye image.
            ProjectAndDraw(theD3DDevice, EyeRight);
        }

        D3DTRY(theD3DDevice->EndScene())

        SAFE_RELEASE(theD3DDevice)

        if (gNumGroupElements == 0)
            D3DTRY(gD3DXContext->DrawDebugText(0.1f, 0.1f, 0x00FFFF, "Choose Instructions... from the Help menu."))

        theDD = gD3DXContext->GetDD();
        if (theDD != NULL)
        {
//          Waiting for the vertical blank is fine when a program
//          is running alone, but when two programs share the screen
//          it creates an artificial bottleneck.
//
//          theDD->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN, NULL);

            SAFE_RELEASE(theDD);
        }

        D3DTRY(gD3DXContext->UpdateFrame(0))
    }
    catch (CException *anException)
    {
        SAFE_DELETE(anException)
        if (theD3DDevice != NULL)
            theD3DDevice->EndScene();
        SAFE_RELEASE(theD3DDevice)

        //  The most likely reason we ended up here is that
        //  we lost our surfaces while the computer was on Stand By,
        //  or when the monitor settings were changed.
        HandleWindowedModeChanges(gD3DXContext);
    }
    catch (...)
    {
        MessageBox(NULL, "Unknown exception -- couldn't render scene.", "Rendering Error", MB_OK);
        if (theD3DDevice != NULL)
            theD3DDevice->EndScene();
        SAFE_RELEASE(theD3DDevice)
    }
}


static void DrawSpaceshipBackground(LPDIRECT3DDEVICE7 aD3DDevice)
{
    //  In a spherical space we see ourselves in all directions,
    //  so the back of our own spaceship fills the sky.
    //  No matter where we move, the image of our own spaceship
    //  is always the same, so draw a precomputed image instead
    //  of rendering the spaceship in real time.

    static D3DXMATRIX
        theDefaultProjection = D3DXMATRIX(
            1, 0,     0, 0,
            0, 1,     0, 0,
            0, 0,  0.5f, 1,
            0, 0, -0.1f, 0),
        theIdentity = D3DXMATRIX(
            1, 0, 0, 0,
            0, 1, 0, 0,
            0, 0, 1, 0,
            0, 0, 0, 1);

    static D3DLVERTEX
        theBackgroundFoggy[4] =
        {
            D3DLVERTEX(D3DVECTOR(-1, -1, 1), 0xFFFFFFFF, SPHERICAL_BACKGROUND_FOG, 0.0f, 1.0f),
            D3DLVERTEX(D3DVECTOR(-1, +1, 1), 0xFFFFFFFF, SPHERICAL_BACKGROUND_FOG, 0.0f, 0.0f),
            D3DLVERTEX(D3DVECTOR(+1, +1, 1), 0xFFFFFFFF, SPHERICAL_BACKGROUND_FOG, 1.0f, 0.0f),
            D3DLVERTEX(D3DVECTOR(+1, -1, 1), 0xFFFFFFFF, SPHERICAL_BACKGROUND_FOG, 1.0f, 1.0f)
        },
        theBackgroundClear[4] =
        {
            D3DLVERTEX(D3DVECTOR(-1, -1, 1), 0xFFFFFFFF, 0xFFFFFFFF, 0.0f, 1.0f),
            D3DLVERTEX(D3DVECTOR(-1, +1, 1), 0xFFFFFFFF, 0xFFFFFFFF, 0.0f, 0.0f),
            D3DLVERTEX(D3DVECTOR(+1, +1, 1), 0xFFFFFFFF, 0xFFFFFFFF, 1.0f, 0.0f),
            D3DLVERTEX(D3DVECTOR(+1, -1, 1), 0xFFFFFFFF, 0xFFFFFFFF, 1.0f, 1.0f)
        };

    D3DTRY(aD3DDevice->SetTransform(D3DTRANSFORMSTATE_WORLD,      theIdentity         ))
    D3DTRY(aD3DDevice->SetTransform(D3DTRANSFORMSTATE_VIEW,       theIdentity         ))
    D3DTRY(aD3DDevice->SetTransform(D3DTRANSFORMSTATE_PROJECTION, theDefaultProjection))

    D3DTRY(aD3DDevice->SetTexture(0, gTextures[TextureSpaceship]))

    D3DTRY(aD3DDevice->DrawPrimitive(
                D3DPT_TRIANGLEFAN,
                D3DFVF_LVERTEX,
                gFog? theBackgroundFoggy : theBackgroundClear,
                4,
                0))
}


static void ProjectAndDraw(
    LPDIRECT3DDEVICE7   aD3DDevice,
    EyeType             anEyeType)
{
    //  Standard rendering shows all visible scenery in flat and hyperbolic spaces.
    //  In a spherical space it shows only the front hemisphere, but if
    //  the scenery objects are distributed symmetrically (with each object
    //  having a twin at its antipodal point) nothing is lost by omitting
    //  the back hemisphere.

    if (gSpaceType != SpaceSpherical
     || gRenderBackHemisphere == FALSE)
    {
        //  Draw the scene.
        DrawContents(aD3DDevice, anEyeType, BananaFull);
    }
    else
    {
        //  Render the front hemisphere viewing banana
        //  into the front half of the clipping box (0 <= z <= 1/2).
        DrawContents(aD3DDevice, anEyeType, BananaFront);

        //  Render the back hemisphere viewing banana
        //  into the back half of the clipping box (1/2 <= z <= 1).
        DrawContents(aD3DDevice, anEyeType, BananaBack);
    }
}


static void DrawContents(
    LPDIRECT3DDEVICE7   aD3DDevice,
    EyeType             anEyeType,
    RangeType           aRangeType)
{
    DWORD   theColor;
    int     i;

    //  Set the projection matrix.
    D3DTRY(aD3DDevice->SetTransform(
        D3DTRANSFORMSTATE_PROJECTION,
        &D3DMATRIX(gProjectionMatrix[anEyeType][aRangeType])))

    //  Choose the color.
    switch (anEyeType)
    {
        case EyeOnly:   theColor = 0xFFFFFFFF;  break;  // white
        case EyeLeft:   theColor = 0xFFFF0000;  break;  // red
        case EyeRight:  theColor = 0xFF00FFFF;  break;  // cyan
    }

    //  Draw the scenery.
    for (i = 0; i < gNumGroupElements; i++)
    {
        //  Cull back-facing triangles.
        D3DTRY(aD3DDevice->SetRenderState(
                D3DRENDERSTATE_CULLMODE,
                (gGroupElementParities[i] == gCameraParity) ?
                    D3DCULL_CCW : D3DCULL_CW))

        //  Draw the Earth (and any other scenery the programmer may add).
        if (gViewEarth)
            gSpinningEarth->Render( aD3DDevice,
                                    &gGroupElements[i],
                                    &gCameraPosition,
                                    gTextures[TextureEarthNorthern],
                                    gTextures[TextureEarthSouthern],
                                    theColor,
                                    aRangeType,
                                    &ComputeFogFactor);

        //  Draw the framework of the Dirichlet domain.
        if (gViewBeams)
            gDirichletDomain->RenderBeams(  aD3DDevice,
                                            &gGroupElements[i],
                                            &gCameraPosition,
                                            gTextures[TextureBeam],
                                            theColor,
                                            aRangeType,
                                            &ComputeFogFactor);
    }

    //  Restore the default cull mode.
    D3DTRY(aD3DDevice->SetRenderState(D3DRENDERSTATE_CULLMODE, D3DCULL_CCW))
}


static DWORD ComputeFogFactor(
    float   *u,
    float   *v,
    int     aRangeType, //  really a RangeType, but cast as an int
                        //      so we can pass the parameter through
                        //      external code
    float   *aDistance)
{
    //  This code is designed to produce an easily viewable result.
    //  However, in a "real hypersphere", for example, the curvature
    //  of the space would tend to refocus the light from distant objects,
    //  offsetting the fog effect and possibly making distant objects
    //  brighter than intermediate objects of the same intrinsic luminosity.

    float   theDistance,
            theFogFactor;

    switch (gSpaceType)
    {
        case SpaceSpherical:
            theDistance = (float) (acos(  u[0]*v[0]
                                        + u[1]*v[1]
                                        + u[2]*v[2]
                                        + u[3]*v[3] ));
            switch ((RangeType) aRangeType)
            {
                case BananaFull:
                    theFogFactor = (theDistance / D3DX_PI) * SPHERICAL_FOG_AT_ANTIPODE;
                    break;
                case BananaFront:
                    theFogFactor = (theDistance / D3DX_PI) * SPHERICAL_FOG_AT_ANTIPODE;
                    break;
                case BananaBack:
                    //  Keep in mind here that theDistance is measured
                    //  from the antipodal camera position, not the
                    //  original camera position.
                    theFogFactor = SPHERICAL_FOG_AT_ANTIPODE
                          +  (SPHERICAL_FOG_MAX - SPHERICAL_FOG_AT_ANTIPODE)
                             * (1.0f - (theDistance / D3DX_PI));
                    break;
            }
            break;

        case SpaceFlat:
            theDistance = (float) (sqrt(  (u[0] - v[0])*(u[0] - v[0])
                                        + (u[1] - v[1])*(u[1] - v[1])
                                        + (u[2] - v[2])*(u[2] - v[2]) ));
            if (theDistance <= FLAT_FOG_BEGIN)
                theFogFactor = 0.0f;
            else
            if (theDistance >= FLAT_FOG_END)
                theFogFactor = 1.0f;
            else
                theFogFactor = (theDistance  - FLAT_FOG_BEGIN)
                             / (FLAT_FOG_END - FLAT_FOG_BEGIN);
            break;

        case SpaceHyperbolic:
            theDistance = (float) (ArcCosh( - u[0]*v[0]
                                            - u[1]*v[1]
                                            - u[2]*v[2]
                                            + u[3]*v[3] ));
            if (theDistance <= HYPERBOLIC_FOG_BEGIN)
                theFogFactor = 0.0f;
            else
            if (theDistance >= HYPERBOLIC_FOG_END)
                theFogFactor = 1.0f;
            else
                theFogFactor = (   theDistance     - HYPERBOLIC_FOG_BEGIN)
                             / (HYPERBOLIC_FOG_END - HYPERBOLIC_FOG_BEGIN);
            break;
    }

    if (aDistance != NULL)
    {
        //  Report the distance to the caller.
        *aDistance = theDistance;

        //  In the spherical case, what counts is
        //  the distance from the camera or its antipodal
        //  point, whichever is less.
        if (gSpaceType == SpaceSpherical
         && *aDistance > D3DX_PI/2)
            *aDistance = D3DX_PI - *aDistance;
    }

    //  Direct3D keeps the fog factor in the alpha component
    //  of the specular color, so we will package up our result
    //  to accomodate this bizarre convention.
    //
    //      If theFogFactor == 0.0, return 0xFFFFFFFF.
    //      If theFogFactor == 1.0, return 0x00FFFFFF.
    //      Otherwise interpolate linearly.
    return (((DWORD)((1.0f - theFogFactor) * 0xFF)) << 24) | 0x00FFFFFF;
}

static double ArcCosh(double x)
{
    if (x >= 1.0)
        return log(x + sqrt(x*x - 1.0));
    else
        return 0.0;
}


static void HandleWindowedModeChanges(ID3DXContext *aD3DXContext)
{
    //  Lots of things can go wrong when different programs
    //  share a graphics card.

    LPDIRECTDRAW7   theDD;

    theDD = aD3DXContext->GetDD();

    //  If some other program grabbed the video card
    //  in exclusive mode, sit tight until it's done.
    while (theDD->TestCooperativeLevel() == DDERR_EXCLUSIVEMODEALREADYSET)
        Sleep(500);

    SAFE_RELEASE(theDD)

    //  Rather than trying to figure out the most efficient
    //  solution for each of the many possible cases,
    //  let's just trash our D3DX Context and start over.
    RestartGraphics();
}


static void PrependTranslation(
    D3DXMATRIX  *aMatrix,
    int         anAxis,
    double      aDistance)
{
    //  Premultiply aMatrix by a translation through aDistance
    //  in the direction of anAxis.

    D3DXMATRIX  theTranslation;

    D3DXMatrixIdentity(&theTranslation);

    switch (gSpaceType)
    {
        case SpaceSpherical:
            theTranslation.m[anAxis][anAxis]    = (float)  cos(aDistance);
            theTranslation.m[3][3]              = (float)  cos(aDistance);
            theTranslation.m[anAxis][3]         = (float) -sin(aDistance);
            theTranslation.m[3][anAxis]         = (float)  sin(aDistance);
            break;

        case SpaceFlat:
            theTranslation.m[3][anAxis]         = (float) aDistance;
            break;

        case SpaceHyperbolic:
            theTranslation.m[anAxis][anAxis]    = (float) cosh(aDistance);
            theTranslation.m[3][3]              = (float) cosh(aDistance);
            theTranslation.m[anAxis][3]         = (float) sinh(aDistance);
            theTranslation.m[3][anAxis]         = (float) sinh(aDistance);
            break;
    }

    D3DXMatrixMultiply( aMatrix,
                        &theTranslation,
                        aMatrix);
}


static void PrependRotation(
    D3DXMATRIX  *aMatrix,
    float       aYaw,
    float       aPitch,
    float       aRoll)
{
    D3DXMATRIX  temp;

    D3DXMatrixMultiply( aMatrix,
                        D3DXMatrixRotationYawPitchRoll(
                            &temp,
                            CAMERA_ROTATION_RATE * aYaw,
                            CAMERA_ROTATION_RATE * aPitch,
                            CAMERA_ROTATION_RATE * aRoll),
                        aMatrix);
}


static void InterpretError(
    HRESULT hr)
{
    char    theErrorString[256];

    D3DXGetErrorString(hr, 256, theErrorString);
    MessageBox(NULL, theErrorString, "D3DX Error", MB_OK);
}


void CheckD3DResult(HRESULT hr)
{
    if FAILED(hr)
    {
        char    *theMessage;

        theMessage = new char[256];
        D3DXGetErrorString(hr, 256, theMessage);
        throw new CException(theMessage, TRUE);
    }
}
