//  CurvedSpaces-Complete-GL.cpp
//
//  This sample code accompanies the article
//
//      "Real-time rendering in curved spaces",
//      IEEE Computer Graphics and Applications, Nov-Dec 2002.
//
//  Interactive 3D animations in spherical, flat, and hyperbolic
//  spaces run in hardware on the user's 3D graphics card via OpenGL.
//
//  Different versions of this code serve different purposes.
//
//      CurvedSpaces-Simple-[D3D or GL].cpp
//
//          This stripped-down version of the program introduces
//          basic techniques of rendering in curved spaces.
//
//      CurvedSpaces-Complete-[D3D or GL].cpp
//
//          This more complete version of the program lets
//          the user explore a variety of 3-dimensional spaces,
//          and may be of interest to geometers and cosmologists
//          as well as computer graphics developers.
//
//  Readers may wish to run the richer CurvedSpaces-Complete.exe
//  to get a feel for what is possible, but read the source code
//  for the more modest CurvedSpaces-Simple.cpp to understand
//  the basics of the implementation.
//
//  Jeff Weeks
//  weeks@geometrygames.org
//  March 2004

//  This code requires the import libraries
//
//      opengl32.lib  user32.lib  gdi32.lib  comdlg32.lib 

//  Known issues:
//
//  (1) OpenGL does not permit assigning explicit fog factors
//      to vertices, but offers only preset flat-space fog formulas.
//      Its three formulas (GL_LINEAR, GL_EXP, GL_EXP2) all use
//      the camera space z-value, and are therefore unsatisfactory
//      for spherical rendering, because in a spherical space the
//      camera space z-value returns to 0 at the antipode.
//      As a workaround, this code uses disables lighting and
//      calls glColor*() to use the current color to fade the scene
//      to black, effectively simulating black fog.
//      An alternative approach would be to enable lighting,
//      use the ambient light level to simulate fog, and use
//      the material's ambient color for some other purpose
//      (such as directional lighting along Clifford parallels
//      as in the CAVE version of this code).
//      The Direct3D version of this program provides colored fog.
//      [There may now be an extension to OpenGL that allows
//      explicitly setting the fog factor at each vertex.]
//
//  (2) As explained in the article, OpenGL's clipping conventions
//      imply that only objects satisfying w > 0 in projection space
//      will be drawn.  However, the OpenGL Red Book cautions that
//      some OpenGL implementations might fail to comply with this rule.
//      I suspect the problem would arise only on older systems,
//      but if this is a problem on your system, uncomment the call
//      to glClipPlane() in the code below to enable an explicit z < 0
//      clipping plane.
//
//  (3) Direct3D and OpenGL use different coordinate systems.
//      By default Direct3D looks in the positive z-direction while
//      OpenGL looks in the negative z-direction.  Thus corrections
//      have been made to prevent the scene viewed in the OpenGL version
//      from being the mirror image of the scene viewed in the Direct3D version.
//      (Note:  This is mostly a "documentation convention" rather
//      than a "hardware convention" or "API convention".  If you avoid
//      glFrustum and set your projection matrix by hand, you can
//      use any coordinate system you want.)
//
//  (4) The spinning Earth's northern hemisphere texture and
//      southern hemisphere texture don't quite match.
//      Most likely this is a problem with how the coordinates
//      got rounded when making the textures, and how they
//      get applied to the triangulation (clamping to the edge
//      probably isn't as good as providing a boundary would be,
//      but I want to keep the textures compatible with both
//      the Direct3D and OpenGL versions of the program).
//
//  (5) For some unknown reason, on my graphics card, OpenGL
//      is fast if and only if you start in 32-bit graphics.
//      Switching to 16-bit graphics once it's running is fine.


#include "Macros-GL.h"
#include "MatrixAlgebra-GL.h"
#include "DirichletDomain-GL.hpp"
#include "SpinningEarth-GL.hpp"
#include "CurvedSpacesResourceIDs-Complete-GL.h"
#include <direct.h> //  for _getcwd()
#include <stdio.h>  //  for file I/O
#include <math.h>

//  Use OpenGL 1.2 feature GL_CLAMP_TO_EDGE if available.
#define GL_CLAMP_TO_EDGE    0x812F

//  How many pixels of empty screen space should
//  we allow around the default window position?
#define WINDOW_MARGIN   16

//  What is the minimum acceptable size for the
//  window's client area?
#define MIN_CLIENT_SIZE 192

//  Pressing the up-arrow key increases the camera speed
//  by CAMERA_SPEED_INCREMENT.  Pressing the down-arrow key
//  decreases it by that amount.  Pressing the space bar
//  sets the camera speed to zero.
#define CAMERA_SPEED_INCREMENT  0.02f

//  After clicking into the scene, moving the mouse pitches and yaws
//  the camera at a rate proportional to CAMERA_ROTATION_RATE.
//  Holding the Ctrl key down while moving the mouse rolls the camera.
#define CAMERA_ROTATION_RATE    1.0f

//  The CLIPPING_DISTANCE should be as close to the camera
//  as possible without inducing z-buffer related artifacts.
//  In a scene of depth ~1, a clipping distance of 0.005f
//  or greater works fine with an ordinary 16-bit depth buffer.
//  [Exercise for advanced readers:  Why would w-buffering fail
//  in spherical rendering?]
#define CLIPPING_DISTANCE   0.01f

//  How closely must the entries match for two matrices
//  to be considered the same?
#define MATRIX_EPSILON      1e-3

//  In a spherical space, objects near the antipodal point
//  appear just as large as nearby objects.  Fog provides
//  an essential depth cue, distinguishing distant large
//  images from nearby ones.  Fog also serves to limit
//  the potentially infinite view in flat and hyperbolic spaces.

//  The flat space fog parameters must be chosen consistently
//  with the typical distance scale of the scenery.
#define FLAT_FOG_BEGIN          1.5f
#define FLAT_FOG_END            2.0f

//  In hyperbolic and spherical spaces there is a natural unit
//  of distance (the radian) so the fog parameters may be set
//  in absolute terms, without reference to the scenery.
#define HYPERBOLIC_FOG_BEGIN    3.0f    //  1.0f
#define HYPERBOLIC_FOG_END      4.0f    //  2.0f
    
//  Fog in a spherical space requires special care.
//  The given settings spread the fog out over the front hemisphere,
//  and leave the back hemisphere rather dim.
#define SPHERICAL_FOG_AT_ANTIPODE   0.8f
#define SPHERICAL_FOG_MAX           0.9f

//  Precompute the OpenGL fog factor on the background image
//  in a spherical space.
#define SPHERICAL_BACKGROUND_FOG    (1.0f - SPHERICAL_FOG_MAX)

//  When testing for fixed points, how close must det(g - I)
//  be to zero to be considered singular?  Matrices with
//  fixed points have det(g - I) very small, so choose
//  SINGULAR_EPSILON large enough to avoid falsely rejecting
//  groups with short translation distances.
#define SINGULAR_EPSILON    1e-8

//  A useful constant.
#define PI  (float)3.14159265358979323846

//  Mouse wheel support is available in Windows ME and later.
#ifndef WM_MOUSEWHEEL
#define WM_MOUSEWHEEL   0x020A
#endif


//  Basic curved space rendering proceeds without knowing
//  the curvature of space.  The exceptions are in
//
//  (1) MatrixTranslation() which must know whether to create
//      a spherical, flat, or hyperbolic rotation matrix when
//      translating the camera or scenery position,
//
//  (2) NormalizeViewMatrix() which maintains an orthogonal
//      matrix for gCameraPosition, and
//
//  (3) Render() which in the spherical case must draw the back
//      hemisphere separately from the front hemisphere, as
//      explained in the accompanying article.
//
//  In CurvedSpacesSimple.cpp these are the *only* uses of
//  the SpaceType.  In CurvedSpacesComplete.cpp the SpaceType
//  is also taken into account when computing fog effects and
//  when constructing certain types of scenery.
typedef enum
{
    SpaceSpherical = 0,
    SpaceFlat,
    SpaceHyperbolic
} SpaceType;

//  In some flat and hyperbolic spaces it's possible to take
//  a trip and return as one's own mirror image, just as a
//  2-dimensional traveller on a Moebius strip returns as her
//  own mirror image.  In these cases we'll need to keep track
//  of whether the camera is mirror reversed or not.
typedef enum
{
    ParityNormal,
    ParityReversed
} ParityType;


//  The possible projection matrices are indexed according
//  to which eye is viewing the scene and whether we are mapping
//  the viewing banana into the front half of the clipping box
//  in projection space, into the back half, or into the whole box. 
typedef enum
{
    EyeOnly = 0,
    EyeLeft,
    EyeRight
} EyeType;
typedef enum
{
    BananaFull,
    BananaFront,
    BananaBack
} RangeType;


//  Different kinds of scenery will require different textures.
//  Be sure to keep this typedef synchronized with gTextureFiles below.
typedef enum
{
    TextureEarthNorthern = 0,
    TextureEarthSouthern,
    TextureBeam,
    TextureSpaceship,
    NumTextures
} TextureType;

//  Win32 bitmaps will provide the textures.
typedef struct
{
    int     width,
            height,
            depth;
    BYTE    *bits;
} BitmapData;

//  Keep Matrices on a linked list while producing the full
//  set of matrices from a set of generators.
typedef struct MatrixNode
{
    Matrix              mm;
    struct MatrixNode   *next;
} MatrixNode;


//  Is fog enabled?
static BOOL gFog = TRUE;

//  Is stereoscopic 3D enabled?
static BOOL gStereo = FALSE;

//  In principle your line of sight in a hypersphere may reach across
//  the hemisphere in front of you, pass through your antipodal point,
//  and arrive at an object in the hemisphere behind you.  In practice
//  this won't happen if the scenery is symmetrically arranged,
//  with every scenery object paired with an identical, antipodal
//  scenery object.  In such cases we can omit the back hemisphere
//  and double the frame rate.
static BOOL gRenderBackHemisphere = FALSE;

//  Show the frame rate in the window title?
static BOOL gShowFrameRate = FALSE;

//  Is the current space spherical, flat, or hyperbolic?
static SpaceType gSpaceType = SpaceFlat;

//  A matrix gCameraPosition would keep track of the camera's
//  position in world space.  Its inverse, gViewMatrix, keeps
//  track of the world's position in camera space.
//  Also keep track of whether the camera is mirror-reversed.
static Matrix           gViewMatrix         = {{
                                                {1,0,0,0},
                                                {0,1,0,0},
                                                {0,0,1,0},
                                                {0,0,0,1}
                                              }};
static float            gCameraSpeed        = 0.0f;
static ParityType       gCameraParity       = ParityNormal;

//  When simulating a manifold, keep a list of all group elements
//  close to the origin.  In the spherical case, keep track of *all*
//  group elements, because there are only finitely many. 
static int              gNumGroupElements       = 0;
static Matrix           *gGroupElements         = NULL;
static ParityType       *gGroupElementParities  = NULL;

//  This sample code uses only one scenery object, namely a
//  spinning earth.  A more interesting program -- say a game
//  in a hypersphere -- would use a variety of scenery objects.
//  CurvedSpaces-Simple-GL.cpp uses a SceneryObject class that may
//  be useful in more substantial programs.
static CSpinningEarth   *gSpinningEarth = NULL;

//  Different kinds of scenery will require different textures.
//  Be sure to keep this typedef synchronized with TextureType above.
static char *gTextureFiles[NumTextures] =
{
    "Textures\\EarthNorthern.bmp",
    "Textures\\EarthSouthern.bmp",
    "Textures\\Beam.bmp",
    "Textures\\Spaceship.bmp"
};

//  For a realistic simulation, the program's internal parameters
//  must match the relative proportions of the user's physical display device,
//  the distance from the display to the user's face, and the separation
//  between the user's eyes.
//  Initialize the parameters for a 90 degree field of view.
//  In monoscopic 3D this gives a nice wide-angle view.
//  In stereoscopic 3D a 90 degree field of view will be practical
//  only with a very large display -- for example in a lecture hall --
//  relative to which the separation between the user's eyes is small.
//  Accurate stereoscopic 3D is more difficult on an ordinary monitor,
//  but is workable in the flat space case if you set gWindowWidth and
//  gWindowDist correctly, but underreport the separation between your eyes.
//  When the user activates Stereo at runtime, he or she is given
//  an opportunity to override the defaults given here.
static float    gWindowWidth    = 200.0f,
                gWindowDist     = 100.0f,
                gEyeSpacing     =   5.0f,
                gIntrinsicUnit  = 400.0f;   //  typically 2*gWindowWidth

//  The nine possible projection matrices are indexed
//  as gProjectionMatrix[EyeType][RangeType].
static Matrix       gProjectionMatrix[3][3];

//  The program uses a single window.
static HWND     gWindow             = NULL;
static HDC      gDeviceContext      = NULL;
static HGLRC    gRenderingContext   = NULL;

//  A Dirichlet domain provides a framework in which to understand
//  the shape of a manifold.
static CDirichletDomain     *gDirichletDomain   = NULL;

//  The group elements defining a manifold are read from a file.
static char         gFileName[MAX_PATH + 1] = "";
static OPENFILENAME gFileInfo = {sizeof(OPENFILENAME), NULL, NULL,
    "Generator files (*.gen)\0*.gen\0Matrix files (*.mat)\0*.mat\0",
    NULL, 0, 0, NULL, MAX_PATH, NULL, 0, NULL, NULL, 0, 0, 0,
    "gen", 0L, NULL, NULL};

//  We'll want to return to our home directory when reading texture files.
static char         gHomeDirectory[256];

//  Viewing options.
static BOOL gViewEarth  = TRUE,
            gViewBeams  = TRUE;

//  Available OpenGL extensions.
static BOOL gHasClampToEdge = FALSE;


static void             InitProjectionMatrices(void);
static void             BuildWindow(HINSTANCE hInstance);
static void             RestartGraphics(void);
static BOOL             InitGraphics(void);
static void             InitOpenGLExtensions(void);
static BOOL             HasOpenGLExtension(char *aDesiredExtensionName);
static void             LoadTextures(void);
static void             LoadBitmapData(BitmapData *aBitmapData, char *aFileName);
static void             FreeBitmapData(BitmapData *aBitmapData);
static void             ShutDownGraphics(void);
static LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
static void             MenuCommand(WORD aCommand, HWND aParentWindow);
static BOOL FAR PASCAL  TrivialDialogProc(HWND hDlg, UINT msg, WPARAM wparam, LPARAM lParam);
static BOOL FAR PASCAL  EnvironmentDialogProc(HWND hDlg, UINT msg, WPARAM wparam, LPARAM lParam);
static void             ReadFile(char *aFileName);
static void             ReadGeneratorFile(char *aFileName);
static BOOL             SameMatrix(Matrix *a, Matrix *b);
static void             ReadMatrixFile(char *aFileName);
static void             Update(void);
static void             NormalizeViewMatrix(void);
static void             Render(void);
static void             DrawSpaceshipBackground(void);
static void             ProjectAndDraw(EyeType anEyeType);
static void             DrawContents(EyeType anEyeType, RangeType aRangeType);
static float            ComputeFogFactor(float *u, float *v, int aRangeType, float *aDistance);
static double           ArcCosh(double x);
static void             MatrixTranslation(Matrix *aMatrix, int anAxis, double aDistance);
static void             PrependTranslation(Matrix *aMatrix, int anAxis, double aDistance);
static void             AppendTranslation(Matrix *aMatrix, int anAxis, double aDistance);
static void             AppendRotation(Matrix *aMatrix, float aPitch, float aYaw, float aRoll);


int PASCAL WinMain(
    HINSTANCE   hInstance,
    HINSTANCE   hPrevInstance, 
    LPSTR       lpszCmdLine,
    int         nCmdShow) 
{
    MSG     theMessage;

    //  Initialize the SpinningEarth.
    gSpinningEarth = new CSpinningEarth();
    if (gSpinningEarth == NULL)
        goto CleanUpAndExit;

    //  Note our home directory, for later use in finding textures.
    _getcwd(gHomeDirectory, 255);

    //  Initialize the window.
    BuildWindow(hInstance);

    //  Initialize the OpenGL graphics.
    if (InitGraphics() == FALSE)
        goto CleanUpAndExit;

    //  Report which version of OpenGL we're using.
//  MessageBox(NULL, (char *)glGetString(GL_VERSION),    "OpenGL Version",    MB_OK);
//  MessageBox(NULL, (char *)glGetString(GL_VENDOR),     "OpenGL Vendor",     MB_OK);
//  MessageBox(NULL, (char *)glGetString(GL_RENDERER),   "OpenGL Renderer",   MB_OK);
//  MessageBox(NULL, (char *)glGetString(GL_EXTENSIONS), "OpenGL Extensions", MB_OK);

    //  Process messages until WM_QUIT is received.
    //  Render the scene during idle time.
    while (TRUE)
    {
        if (PeekMessage(&theMessage, NULL, 0, 0, PM_REMOVE) != 0)
        {
            TranslateMessage(&theMessage);
            DispatchMessage(&theMessage);

            if (theMessage.message == WM_QUIT)
                break;
        }
        else
        {
            if (IsIconic(gWindow) == FALSE
             && gNumGroupElements > 0)
            {
                Update();
                Render();
			}
            else
                Sleep(100);
        }
    }

CleanUpAndExit:

    //  Shut down OpenGL.
    ShutDownGraphics();

    //  Release the window.
    if (gWindow != NULL)
        DestroyWindow(gWindow);

    //  Delete the SpinningEarth.
    SAFE_DELETE(gSpinningEarth)

    //  Delete the GroupElements.
    SAFE_DELETE_(gGroupElements);
    SAFE_DELETE_(gGroupElementParities);

    //  Delete the Dirichlet domain.
    SAFE_DELETE(gDirichletDomain);

    return 0;
}


static void InitProjectionMatrices(void)
{
    RECT    theClientRect;
    float   theWindowHeight;
    int     i,
            j,
            k,
            l;

    //  Deduce theWindowHeight from the aspect ratio
    //  of gWindow's client area.
    if (gWindow != NULL)
    {
        GetClientRect(gWindow, &theClientRect);
        theWindowHeight = (gWindowWidth * theClientRect.bottom) / theClientRect.right;
    }
    else
        theWindowHeight = gWindowWidth;


    //  Start with standard banana matrices.  By default the program
    //  begins with aWindowWidth == 2*aWindowDistance, giving a nice
    //  wide 90 degree field of view.  It's easy to verify that the
    //  standard banana matrix takes the corners of the viewing banana
    //  to the corners of the clipping box in projection space.
    //  Please see the accompanying article "Real-Time Rendering
    //  in Curved Spaces" for details.
    for (i = 0; i < 3; i++)
        for (j = 0; j < 3; j++)
        {
            MatrixIdentity(&gProjectionMatrix[i][j]);

            gProjectionMatrix[i][j].m[0][0] = 2 * gWindowDist / gWindowWidth;
            gProjectionMatrix[i][j].m[1][1] = 2 * gWindowDist / theWindowHeight;

            gProjectionMatrix[i][j].m[2][2] =  0.0f;
            gProjectionMatrix[i][j].m[2][3] = -1.0f;
            gProjectionMatrix[i][j].m[3][2] = -CLIPPING_DISTANCE;
            gProjectionMatrix[i][j].m[3][3] =  0.0f;
        }

    //  Modify the BananaFront and BananaBack matrices to map into
    //  the front and back halves of the clipping box, respectively.
    //  Again, please see the main article for a complete explanation.
    for (i = 0; i < 3; i++)
    {
        gProjectionMatrix[i][BananaFront].m[2][2] =  0.5f;
        gProjectionMatrix[i][BananaFront].m[3][2] = -0.5f * CLIPPING_DISTANCE;

        gProjectionMatrix[i][BananaBack ].m[2][2] = -0.5f;
        gProjectionMatrix[i][BananaBack ].m[3][2] = -0.5f * CLIPPING_DISTANCE;
    }

    //  Offset the EyeLeft and EyeRight projections, to account for
    //  the fact that the left eye sees the window a little to its right,
    //  and the right eye sees it a little to its left.  The guiding
    //  principle is that a person using the program should receive
    //  the same light from the same directions as would a person
    //  in the actual spherical, flat, or hyperbolic space itself.
    for (j = 0; j < 3; j++)
    {
        //  2002/06/11  Corrected the signs to be the opposite of those
        //  in the Direct3D version of this program, to allow for the fact
        //  that the z coordinate will be negative.  Stereoscopic 3D now works
        //  correctly in this OpenGL version as well as in the Direct3D version.
        gProjectionMatrix[EyeLeft ][j].m[2][0] = + gEyeSpacing / gWindowWidth;
        gProjectionMatrix[EyeRight][j].m[2][0] = - gEyeSpacing / gWindowWidth;
    }

    //  Multiply the BananaBack matrices by -Id to invert the scenery
    //  and eliminate the need to invert the camera at render time.
    for (i = 0; i < 3; i++)
        for (k = 0; k < 4; k++)
            for (l = 0; l < 4; l++)
                gProjectionMatrix[i][BananaBack].m[k][l] *= -1;

    //  Strictly speaking, the projection matrices are now complete.
    //  However, to save ourselves the trouble of offsetting the camera
    //  for the EyeLeft and EyeRight stereo views at render time, we can
    //  instead offset the scenery, and pre-multiply the offset into the
    //  projection matrices.
    for (j = 0; j < 3; j++)
    {
        PrependTranslation( &gProjectionMatrix[EyeLeft][j],
                            0,
                            + (gEyeSpacing / gIntrinsicUnit) / 2);

        PrependTranslation( &gProjectionMatrix[EyeRight][j],
                            0,
                            - (gEyeSpacing / gIntrinsicUnit) / 2);
    }
}


static void BuildWindow(
    HINSTANCE   hInstance)
{
    int theClientWidth,
        theClientHeight;

    //  Register the window class.
    WNDCLASS wndClass =
    {
        CS_HREDRAW | CS_VREDRAW,
        WndProc,
        0,
        0,
        hInstance,
        LoadIcon(hInstance, "CurvedSpacesCompleteIcon"),
        LoadCursor(hInstance, "CurvedSpacesCursor"),
        (HBRUSH)GetStockObject(BLACK_BRUSH),
        TEXT("CurvedSpacesMenu"),
        TEXT("Curved Spaces Window Class")
    };
    RegisterClass(&wndClass);

    //  Figure out the largest size square window
    //  that will fit comfortably on the screen.
    theClientWidth  =     GetSystemMetrics(SM_CXSCREEN)
                    - 2 * GetSystemMetrics(SM_CXSIZEFRAME)
                    - 2 * WINDOW_MARGIN;
    theClientHeight =     GetSystemMetrics(SM_CYSCREEN)
                    -     GetSystemMetrics(SM_CYCAPTION)
                    -     GetSystemMetrics(SM_CYMENU)
                    - 2 * GetSystemMetrics(SM_CYSIZEFRAME)
                    - 2 * WINDOW_MARGIN;
    if (theClientWidth > theClientHeight)
        theClientWidth = theClientHeight;
    if (theClientHeight > theClientWidth)
        theClientHeight = theClientWidth;

    //  Create the window.
    gWindow = CreateWindow (
        TEXT("Curved Spaces Window Class"),
        TEXT("Curved Spaces (OpenGL)"),
        WS_OVERLAPPEDWINDOW | WS_VISIBLE,
        WINDOW_MARGIN, WINDOW_MARGIN,
        theClientWidth  + 2 * GetSystemMetrics(SM_CXSIZEFRAME),
        theClientHeight + GetSystemMetrics(SM_CYCAPTION)
                        + GetSystemMetrics(SM_CYMENU)
                        + 2 * GetSystemMetrics(SM_CYSIZEFRAME),
        NULL, NULL, hInstance, NULL);
}



static void RestartGraphics(void)
{
    if (IsIconic(gWindow))
        return;

    ShutDownGraphics();

    if (InitGraphics() == FALSE)
    {
        MessageBox(NULL, "Can't reinitialize graphics.  Bye.", "We're cooked.", MB_OK);
        PostQuitMessage(0);
    }
}


static BOOL InitGraphics(void)
{
    GLuint  thePixelFormat;

    static PIXELFORMATDESCRIPTOR    thePixelFormatDescriptor =
    {
        sizeof(PIXELFORMATDESCRIPTOR),  //  size
        1,                              //  version number
        PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER, //  format
        PFD_TYPE_RGBA,                  //  preferred format is rgba (not palette)
        24,                             //  preferred depth
        0, 0, 0, 0, 0, 0,               //  color bits ignored
        8,                              //  alpha buffer depth
        0,                              //  shift bit ignored
        0,                              //  no accumulation buffer
        0, 0, 0, 0,                     //  accumulation bits ignored
        16,                             //  16-bit Z-buffer  
        0,                              //  no stencil buffer
        0,                              //  no auxiliary buffer
        0,                              //  ignored
        PFD_MAIN_PLANE,                 //  underlay/overlay planes
        0,                              //  ignored
        0,                              //  underlay color
        0                               //  ignored
    };

    gDeviceContext = GetDC(gWindow);
    if (gDeviceContext == NULL)
    {
        MessageBox(NULL, "Can't get a device context to initialize OpenGL.", "GetDC() failed", MB_OK);
        return FALSE;
    }

    thePixelFormat = ChoosePixelFormat(gDeviceContext, &thePixelFormatDescriptor);
    if (thePixelFormat == 0)
    {
        MessageBox(NULL, "Can't get a pixel format to initialize OpenGL.", "ChoosePixelFormat() failed", MB_OK);
        return FALSE;
    }

    if (SetPixelFormat(gDeviceContext, thePixelFormat, &thePixelFormatDescriptor) == FALSE)
    {
        MessageBox(NULL, "Can't set the pixel format to initialize OpenGL.", "SetPixelFormat() failed", MB_OK);
        return FALSE;
    }

    gRenderingContext = wglCreateContext(gDeviceContext);
    if (gRenderingContext == NULL)
    {
        MessageBox(NULL, "Can't get an OpenGL rendering context.", "wglCreateContext() failed", MB_OK);
        return FALSE;
    }

    if (wglMakeCurrent(gDeviceContext, gRenderingContext) == FALSE)
    {
        MessageBox(NULL, "Can't set the current rendering context.", "wglMakeCurrent() failed", MB_OK);
        return FALSE;
    }

    InitOpenGLExtensions();

    LoadTextures();

    if (gSpinningEarth != NULL)
        gSpinningEarth->MakeDisplayLists(TextureEarthNorthern, TextureEarthSouthern);

    return TRUE;
}


static void InitOpenGLExtensions(void)
{
//  BOOL    (WINAPI *theSwapIntervalEXT)(int interval);

    //  Is GL_CLAMP_TO_EDGE available?
    gHasClampToEdge = HasOpenGLExtension("GL_EXT_texture_edge_clamp ");

    //  If wglSwapIntervalEXT is available,
    //  ask it to wait for the vertical refresh.
    //  (Note:  We don't really need to call HasOpenGLExtension() here.)
//  if (HasOpenGLExtension("WGL_EXT_swap_control "))
//  {
//      Waiting for the vertical refresh is fine when a program
//      is running alone, but when two programs share the screen
//      it creates an artificial bottleneck.
//
//      theSwapIntervalEXT = (BOOL (WINAPI *)(int)) wglGetProcAddress("wglSwapIntervalEXT");
//      if (theSwapIntervalEXT != NULL)
//          theSwapIntervalEXT(1);
//  }
}


static BOOL HasOpenGLExtension(
    char    *aDesiredExtensionName)
{
    //  IMPORTANT:  Make sure aDesiredExtensionName includes
    //  a trailing space, to avoid false matches.
    //  A search for "foo" might find a false match
    //  in "foobar", but searching for "foo " is safe
    //  even in the presence of "foobar ". 
    
    char    *theExtensionList;

    theExtensionList = (char *) glGetString(GL_EXTENSIONS);
    if (theExtensionList == NULL)
        return FALSE;

    return (strstr(theExtensionList, aDesiredExtensionName) != NULL);
}


static void LoadTextures(void)
{
    BitmapData  theBitmapData;
    int         i;

    //  Look for the texture files relative to the home directory.
    _chdir(gHomeDirectory);

    //  Load any desired textures.
    //  If we can't load the textures, push on without them.
    for (i = 0; i < NumTextures; i++)
    {
        LoadBitmapData(&theBitmapData, gTextureFiles[i]);

        if (theBitmapData.bits != NULL)
        {
            glBindTexture(GL_TEXTURE_2D, i+1);

            if (theBitmapData.depth == 24)
                glTexImage2D(   GL_TEXTURE_2D,
                                0,
                                GL_RGBA,
                                theBitmapData.width,
                                theBitmapData.height,
                                0,
                                GL_RGB,
                                GL_UNSIGNED_BYTE,
                                theBitmapData.bits);
            else
                glTexImage2D(   GL_TEXTURE_2D,
                                0,
                                GL_LUMINANCE,
                                theBitmapData.width,
                                theBitmapData.height,
                                0,
                                GL_LUMINANCE,
                                GL_UNSIGNED_BYTE,
                                theBitmapData.bits);


            //  Set minification and magnification filters.
            //  (This step is crucial -- the default min filter won't work!)
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

            //  Wrap texture coordinates as needed.
            switch (i)
            {
                case TextureEarthNorthern:
                case TextureEarthSouthern:
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, gHasClampToEdge ? GL_CLAMP_TO_EDGE : GL_CLAMP);
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, gHasClampToEdge ? GL_CLAMP_TO_EDGE : GL_CLAMP);
                    break;

                case TextureBeam:
                case TextureSpaceship:
                default:
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
                    break;
            }
        }

        FreeBitmapData(&theBitmapData);
    }
}


static void LoadBitmapData(
    BitmapData  *aBitmapData,
    char        *aFileName)
{
    HANDLE              theFile;
    BITMAPFILEHEADER    theFileHeader;
    char                *theBitmapInfo;
    HBITMAP             theBitmap;
    BYTE                *theBits;
    DWORD               theNumBytesRead,
                        theBitmapInfoSize,
                        theNumBytes;
    int                 i,
                        j,
                        ii;
    BOOL                theResult;
    char                theBuffer[512];

    theFile             = INVALID_HANDLE_VALUE;
    theBitmapInfo       = NULL;
    theBitmap           = NULL;
    aBitmapData->width  = 0;
    aBitmapData->height = 0;
    aBitmapData->depth  = 0;
    aBitmapData->bits   = NULL;

    theFile = CreateFile(aFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
    if (theFile == INVALID_HANDLE_VALUE)
        goto CleanupLoadBitmapData;

    theResult = ReadFile(theFile, &theFileHeader, sizeof(BITMAPFILEHEADER), &theNumBytesRead, NULL);
    if (theResult == FALSE
     || theNumBytesRead != sizeof(BITMAPFILEHEADER)
     || theFileHeader.bfType != *(WORD *)"BM")
        goto CleanupLoadBitmapData;

    theBitmapInfoSize = theFileHeader.bfOffBits - sizeof(BITMAPFILEHEADER);
    theBitmapInfo = new char[theBitmapInfoSize];
    if (theBitmapInfo == NULL)
        goto CleanupLoadBitmapData;

    theResult = ReadFile(theFile, theBitmapInfo, theBitmapInfoSize, &theNumBytesRead, NULL);
    if (theResult == FALSE
     || theNumBytesRead != theBitmapInfoSize)
        goto CleanupLoadBitmapData;

    theBitmap = CreateDIBSection(NULL, (BITMAPINFO *)theBitmapInfo, DIB_RGB_COLORS, (void **)&theBits, NULL, 0);
    if (theBitmap == NULL)
        goto CleanupLoadBitmapData;

    theNumBytes = theFileHeader.bfSize - theFileHeader.bfOffBits;
    theResult = ReadFile(theFile, theBits, theNumBytes, &theNumBytesRead, NULL);
    if (theResult == FALSE
     || theNumBytesRead != theNumBytes)
        goto CleanupLoadBitmapData;

    aBitmapData->width  = ((BITMAPINFO *)theBitmapInfo)->bmiHeader.biWidth;
    aBitmapData->height = ((BITMAPINFO *)theBitmapInfo)->bmiHeader.biHeight;
    aBitmapData->depth  = ((BITMAPINFO *)theBitmapInfo)->bmiHeader.biBitCount;

    //  Make a private copy of theBits.

    switch (aBitmapData->depth)
    {
        case 24:    //  color bitmap for mono use

            //  Switch GBR to RGB, and invert the image top-to-bottom.

            aBitmapData->bits = new BYTE[aBitmapData->width * aBitmapData->height * 3];
            if (aBitmapData->bits == NULL)
                goto CleanupLoadBitmapData;

            for (i = 0; i < aBitmapData->height; i++)
            {
                ii = aBitmapData->height - 1 - i;
                for (j = 0; j < aBitmapData->width; j++)
                {
                    aBitmapData->bits[ 3*( i*aBitmapData->width + j) + 0]
                          = theBits[ 3*(ii*aBitmapData->width + j) + 2];

                    aBitmapData->bits[ 3*( i*aBitmapData->width + j) + 1]
                          = theBits[ 3*(ii*aBitmapData->width + j) + 1];

                    aBitmapData->bits[ 3*( i*aBitmapData->width + j) + 2]
                          = theBits[ 3*(ii*aBitmapData->width + j) + 0];
                }
            }
            break;

        case 8:     //  greyscale bitmap for stereo use

            //  Invert the image top-to-bottom.

            aBitmapData->bits = new BYTE[aBitmapData->width * aBitmapData->height * 1];
            if (aBitmapData->bits == NULL)
                goto CleanupLoadBitmapData;

            for (i = 0; i < aBitmapData->height; i++)
            {
                ii = aBitmapData->height - 1 - i;
                for (j = 0; j < aBitmapData->width; j++)
                    aBitmapData->bits[ i*aBitmapData->width + j]
                            = theBits[ii*aBitmapData->width + j];
            }
            break;

        default:
            goto CleanupLoadBitmapData;
    }

CleanupLoadBitmapData:

    if (aBitmapData->bits == NULL)
    {
        wsprintf(theBuffer, "Couldn't load texture from file %s.  Most likely the file is missing or invalid.", aFileName);
        MessageBox(NULL, theBuffer, "Texture Failure", MB_OK);
    }

    if (theBitmap != NULL)
    {
        DeleteObject(theBitmap);    //  automatically frees theBits as well
        theBitmap = NULL;
    }

    SAFE_DELETE_(theBitmapInfo);

    if (theFile != INVALID_HANDLE_VALUE)
        CloseHandle(theFile);
}


static void FreeBitmapData(BitmapData *aBitmapData)
{
    SAFE_DELETE_(aBitmapData->bits);
}


static void ShutDownGraphics(void)
{
    GLuint  i;

    if (gSpinningEarth != NULL)
        gSpinningEarth->FreeDisplayLists();

    glBindTexture(GL_TEXTURE_2D, 0);
    for (i = 1; i <= NumTextures; i++)
        glDeleteTextures(1, &i);

    wglMakeCurrent(NULL, NULL);

    if (gRenderingContext != NULL)
    {
        wglDeleteContext(gRenderingContext);
        gRenderingContext = NULL;
    }

    if (gDeviceContext != NULL)
    {
        ReleaseDC(gWindow, gDeviceContext);
        gDeviceContext = NULL;
    }
}


//  Callback function to process window messages.
static LRESULT CALLBACK WndProc(
    HWND    hWnd,
    UINT    uMsg,
    WPARAM  wParam,
    LPARAM  lParam)
{
    RECT            theRect;
    POINT           theNewCursorLocation;
    float           dx,
                    dy;
    static POINT    theWindowCenter = {0,0};
    static BOOL     theCursorIsHidden = FALSE;

    switch (uMsg)
    {
        case WM_CLOSE:
            PostQuitMessage(0);
            return 0;

        case WM_DESTROY:
            gWindow = NULL;
            return 0;

        case WM_KEYDOWN:
            switch (wParam) //  virtual key code
            {
                case VK_UP:
                    gCameraSpeed += CAMERA_SPEED_INCREMENT;
                    break;

                case VK_DOWN:
                    gCameraSpeed -= CAMERA_SPEED_INCREMENT;
                    break;

                case ' ':
                    gCameraSpeed = 0;
                    break;

                case VK_ESCAPE:
                    //  The user wants to stop flying.
                    //  ReleaseCapture() generates a WM_CAPTURECHANGED message
                    //  which will be handled below.
                    ReleaseCapture();
                    break;
            }
            return 0;

        case WM_COMMAND:
            MenuCommand(LOWORD(wParam), hWnd);
            return 0;

        case WM_LBUTTONDOWN:
            //  The user is clicking into the scene to steer.
            //  Capture the mouse and hide the cursor.
            SetCapture(hWnd);
            SetCursor(NULL);
            GetWindowRect(hWnd, &theRect);
            theWindowCenter.x = (theRect.right  + theRect.left)/2;
            theWindowCenter.y = (theRect.bottom + theRect.top )/2;
            SetCursorPos(theWindowCenter.x, theWindowCenter.y);
            theCursorIsHidden = TRUE;
            return 0;

        case WM_RBUTTONDOWN:
            //  The user wants to stop flying.
            //  ReleaseCapture() generates a WM_CAPTURECHANGED message
            //  which will be handled below.
            ReleaseCapture();
            return 0;

        case WM_CAPTURECHANGED:
            //  The WM_CAPTURECHANGED message typically results
            //  from our own call to ReleaseCapture(),
            //  but it could arise in other ways, for example,
            //  if some other program grabs the mouse.
            theCursorIsHidden = FALSE;
            return 0;

        case WM_MOUSEMOVE:
            //  When the cursor is hidden, mouse movements steer the camera.
            if (theCursorIsHidden)
            {
                GetCursorPos(&theNewCursorLocation);

                if (theNewCursorLocation.x != theWindowCenter.x
                 || theNewCursorLocation.y != theWindowCenter.y)
                {
                    GetWindowRect(hWnd, &theRect);

                    dx = (float)(theNewCursorLocation.x - theWindowCenter.x)
                       / (float)(theRect.right  - theRect.left);
                    dy = (float)(theNewCursorLocation.y - theWindowCenter.y)
                       / (float)(theRect.bottom - theRect.top );

                    SetCursorPos(theWindowCenter.x, theWindowCenter.y);

                    //  Allow full six-degrees-of-freedom navigation.
                    if (wParam & MK_SHIFT)
                    {
                        //  Translate
                        if (wParam & MK_CONTROL)
                            AppendTranslation(&gViewMatrix, 2, -dy);
                        else
                        {
                            AppendTranslation(&gViewMatrix, 0, -dx);
                            AppendTranslation(&gViewMatrix, 1,  dy);
                        }
                    }
                    else
                    {
                        //  Rotate
                        if (wParam & MK_CONTROL)   
                            AppendRotation(&gViewMatrix,  0,  0, dx);                         
                        else
                            AppendRotation(&gViewMatrix, dy, dx,  0);
                    }
                }
            }
            return 0;

        case WM_MOUSEWHEEL:
            gCameraSpeed += CAMERA_SPEED_INCREMENT * ((short)HIWORD(wParam))/120.0f;
            return 0;

        case WM_MBUTTONDOWN:    //  perhaps user has pressed mouse wheel
            gCameraSpeed = 0;
            return 0;

        case WM_SIZE:
            InitProjectionMatrices();
            if (gWindow != NULL)
                glViewport(0, 0, LOWORD(lParam), HIWORD(lParam));
            return 0;

        case WM_GETMINMAXINFO:
            // Prevent the client area from going smaller
            //  than MIN_CLIENT_SIZE.
            ((MINMAXINFO*)lParam)->ptMinTrackSize.x =
                MIN_CLIENT_SIZE
                + 2 * GetSystemMetrics(SM_CXSIZEFRAME);
            ((MINMAXINFO*)lParam)->ptMinTrackSize.y =
                MIN_CLIENT_SIZE
                + GetSystemMetrics(SM_CYCAPTION)
                + GetSystemMetrics(SM_CYMENU)
                + 2 * GetSystemMetrics(SM_CYSIZEFRAME);
            return 0;

        case WM_INITMENUPOPUP:

            CheckMenuItem(  (HMENU)wParam, ID_VIEW_EARTH,
                gViewEarth            ? MF_CHECKED : MF_UNCHECKED);
            CheckMenuItem(  (HMENU)wParam, ID_VIEW_BEAMS,
                gViewBeams            ? MF_CHECKED : MF_UNCHECKED);
            CheckMenuItem(  (HMENU)wParam, ID_OPTIONS_FOG,
                gFog                  ? MF_CHECKED : MF_UNCHECKED);
            CheckMenuItem(  (HMENU)wParam, ID_OPTIONS_STEREO,
                gStereo               ? MF_CHECKED : MF_UNCHECKED);
            CheckMenuItem(  (HMENU)wParam, ID_OPTIONS_BACKSIDE,
                gRenderBackHemisphere ? MF_CHECKED : MF_UNCHECKED);
            CheckMenuItem(  (HMENU)wParam, ID_OPTIONS_FRAMERATE,
                gShowFrameRate        ? MF_CHECKED : MF_UNCHECKED);
 
            return 0;

        default:
            return DefWindowProc(hWnd, uMsg, wParam, lParam);
    }
}


static void MenuCommand(
    WORD    aCommand,
    HWND    aParentWindow)
{
    switch (aCommand)
    {
        //  file menu

        case ID_FILE_OPEN:
            gFileInfo.hwndOwner         = gWindow;
            gFileInfo.lpstrFile         = gFileName;
            gFileInfo.Flags             = 0;
            if (GetOpenFileName(&gFileInfo))
                ReadFile(gFileName);
            break;

        case ID_FILE_EXIT:
            PostQuitMessage(0);
            break;

        //  view menu

        case ID_VIEW_EARTH:
            gViewEarth = ! gViewEarth;
            break;

        case ID_VIEW_BEAMS:
            gViewBeams = ! gViewBeams;
            break;

        //  options menu

        case ID_OPTIONS_FOG:
            gFog = ! gFog;
            break;

        case ID_OPTIONS_STEREO:
            gStereo = ! gStereo;
            break;

        case ID_OPTIONS_PERSPECTIVE:
            DialogBox(  (HINSTANCE)GetWindowLong(aParentWindow, GWL_HINSTANCE),
                        TEXT("EnvironmentDialog"),
                        aParentWindow, 
                        (DLGPROC)EnvironmentDialogProc);
            break;

        case ID_OPTIONS_BACKSIDE:
            gRenderBackHemisphere = ! gRenderBackHemisphere;
            break;

        case ID_OPTIONS_FRAMERATE:
            gShowFrameRate = ! gShowFrameRate;
            SetWindowText(gWindow, gShowFrameRate ? "Counting..." : "Curved Spaces (OpenGL)");
            break;

        //  help menu

        case ID_HELP_INSTRUCTIONS:
            DialogBox(  (HINSTANCE)GetWindowLong(aParentWindow, GWL_HINSTANCE),
                        TEXT("INSTRUCTIONSDIALOG"),
                        aParentWindow, 
                        (DLGPROC)TrivialDialogProc);
            break;

        case ID_HELP_ABOUT:
            DialogBox(  (HINSTANCE)GetWindowLong(aParentWindow, GWL_HINSTANCE),
                        TEXT("ABOUTDIALOG"),
                        aParentWindow, 
                        (DLGPROC)TrivialDialogProc);
            break;
    }
}

static BOOL FAR PASCAL TrivialDialogProc(
    HWND    hDlg,
    UINT    msg,
    WPARAM  wparam,
    LPARAM  lParam)
{
    switch (msg)
    {
        case WM_INITDIALOG:
            return TRUE;

        case WM_COMMAND:
            switch (LOWORD(wparam))
            {
                case IDOK:
                    EndDialog(hDlg, TRUE);
                    return TRUE;

                default:
                    return FALSE;
            }

        default:
            return FALSE;
    }
}

static BOOL FAR PASCAL EnvironmentDialogProc(
    HWND    hDlg,
    UINT    msg,
    WPARAM  wparam,
    LPARAM  lParam)
{
    char    theBuffer[32];
    float   theNumber;

    switch (msg)
    {
        case WM_INITDIALOG:

            //  Copy our internal variables into the dialog's edit fields.

            sprintf(theBuffer, "%.0f", gWindowWidth);
            SetWindowText(GetDlgItem(hDlg, IDC_WINDOW_WIDTH   ), theBuffer);

            sprintf(theBuffer, "%.0f", gWindowDist);
            SetWindowText(GetDlgItem(hDlg, IDC_WINDOW_DISTANCE), theBuffer);

            sprintf(theBuffer, "%.1f", gEyeSpacing);
            SetWindowText(GetDlgItem(hDlg, IDC_EYE_SPACING    ), theBuffer);

            sprintf(theBuffer, "%.0f", gIntrinsicUnit);
            SetWindowText(GetDlgItem(hDlg, IDC_INTRINSIC_UNIT ), theBuffer);

            return TRUE;

        case WM_COMMAND:

            switch (LOWORD(wparam))
            {
                case IDOK:

                    //  Copy the dialog's edit fields back into our internal variables.

                    GetWindowText(GetDlgItem(hDlg, IDC_WINDOW_WIDTH   ), theBuffer, 32);
                    if (sscanf(theBuffer, "%f", &theNumber) == 1)
                        gWindowWidth = theNumber;

                    GetWindowText(GetDlgItem(hDlg, IDC_WINDOW_DISTANCE), theBuffer, 32);
                    if (sscanf(theBuffer, "%f", &theNumber) == 1)
                        gWindowDist = theNumber;

                    GetWindowText(GetDlgItem(hDlg, IDC_EYE_SPACING    ), theBuffer, 32);
                    if (sscanf(theBuffer, "%f", &theNumber) == 1)
                        gEyeSpacing = theNumber;

                    GetWindowText(GetDlgItem(hDlg, IDC_INTRINSIC_UNIT ), theBuffer, 32);
                    if (sscanf(theBuffer, "%f", &theNumber) == 1)
                        gIntrinsicUnit = theNumber;

                    //  Dismiss the dialog.
                    EndDialog(hDlg, TRUE);

                    //  Recompute the projection matrices using the new parameters.
                    InitProjectionMatrices();

                    return TRUE;

                case IDCANCEL:
                    EndDialog(hDlg, TRUE);
                    return TRUE;

                default:
                    return FALSE;
            }
            return TRUE;
    }

    return FALSE;
}


static void ReadFile(char *aFileName)
{
    char    *theFileExtension;

    theFileExtension = aFileName + strlen(aFileName) - 3;

    if (strcmp(theFileExtension, "gen") == 0)
        ReadGeneratorFile(aFileName);

    if (strcmp(theFileExtension, "mat") == 0)
        ReadMatrixFile(aFileName);
}


static void ReadGeneratorFile(char *aFileName)
{
    FILE        *theFile;
    double      theCutoffRadius,
                theTranslationDistance;
    float       theDummy;
    int         theCount,
                theResult,
                theNumGenerators,
                theNumMatrices,
                i,
                j,
                k;
    Matrix      *theGenerators,
                *theMatrices,
                theCandidate;
//  Matrix      theMatrix;
    MatrixNode  *theFirstNode,
                *theSecondNode,
                *theNode,
                *theExistingNode,
                *theLastNode;
    ParityType  *theParities;
    SpaceType   theSpaceType,
                theNewSpaceType;
    BOOL        theCandidateIsOnTheList;
    char        *theWindowTitle,
                *theChar;

    theFile         = NULL;
    theGenerators   = NULL;
    theMatrices     = NULL;
    theFirstNode    = NULL;
    theParities     = NULL;

    SetCursor(LoadCursor(NULL, IDC_WAIT));

    try
    {
        //  Open the file.
        theFile = fopen(aFileName, "r");
        if (theFile == NULL)
            throw new CException("Couldn't open file.");

        //  Read the cutoff radius.
        if (1 != fscanf(theFile, "%lf", &theCutoffRadius))
            throw new CException("Couldn't read cutoff radius.");

        //  Count the total number of matrix entries in the file.
        theCount = 0;
        do
        {
            theResult = fscanf(theFile, "%f", &theDummy);
            if (theResult == 1)
                theCount++;
            else if (theResult != EOF)
                throw new CException("File contains non-numeric entries.");
        } while (theResult == 1);

        //  Reset the file pointer to the first matrix entry.
        rewind(theFile);
        fscanf(theFile, "%lf", &theCutoffRadius);

        //  The number of entries should be a multiple of 16.
        if (theCount%16 != 0)
            throw new CException("File format:  cutoff radius followed by 4x4 generator matrices.");

        //  How many matrices does the file contain?
        theNumGenerators = theCount/16;

        //  Allocate space for the generators and their inverses.
        theGenerators = new Matrix[2*theNumGenerators];
        if (theGenerators == NULL)
            throw new CException("Memory allocation failed.");

        //  Read the generators.
        for (i = 0; i < theNumGenerators; i++)
        {
            for (j = 0; j < 4; j++)
                for (k = 0; k < 4; k++)
                    if (1 != fscanf(theFile, "%f", &theGenerators[2*i].m[j][k]))
                        throw new CException("Error reading generators.");

            //  2002/10/13  Conjugate by the reflection diag(1,1,-1,1)
            //  to compensate for the difference in orientation conventions
            //  between OpenGL and Direct3D.
            for (j = 0; j < 4; j++)
            {
                theGenerators[2*i].m[j][2] = - theGenerators[2*i].m[j][2];
                theGenerators[2*i].m[2][j] = - theGenerators[2*i].m[2][j];
            }

            if (theGenerators[2*i].m[3][3] <  1.0f)
                theNewSpaceType = SpaceSpherical;
            if (theGenerators[2*i].m[3][3] == 1.0f)
                theNewSpaceType = SpaceFlat;
            if (theGenerators[2*i].m[3][3] >  1.0f)
                theNewSpaceType = SpaceHyperbolic;

            if (i == 0)
                theSpaceType = theNewSpaceType;
            else
            {
                if (theSpaceType != theNewSpaceType)
                    throw new CException("Matrices belong to inconsistent geometries. (Or perhaps the identity was unnecessarily included in the file.)");
            }

            //  Compute the generator's inverse.
            switch (theSpaceType)
            {
                case SpaceSpherical:
                    MatrixTranspose(&theGenerators[2*i+1], &theGenerators[2*i]);
                    break;
                case SpaceFlat:
                    MatrixInverse(&theGenerators[2*i+1], &theGenerators[2*i]);
                    break;
                case SpaceHyperbolic:
                    MatrixTranspose(&theGenerators[2*i+1], &theGenerators[2*i]);
                    for (j = 0; j < 4; j++)
                        for (k = 0; k < 4; k++)
                            if ((j == 3) != (k == 3))
                                theGenerators[2*i+1].m[j][k] = - theGenerators[2*i+1].m[j][k];
                    break;
            }
        }

        //  Close the file.
        fclose(theFile);
        theFile = NULL;

        //  Find all group elements out to theCutoffRadius
        //  and put them on a NULL-terminated singly-linked list.

        theFirstNode = new MatrixNode;
        if (theFirstNode == NULL)
            throw new CException("Memory allocation failed.");
        MatrixIdentity(&theFirstNode->mm);
        theFirstNode->next = NULL;
        theLastNode = theFirstNode;
        theNumMatrices = 1;

        for (   theNode = theFirstNode;
                theNode != NULL;
                theNode = theNode->next)
        {
            for (i = 0; i < 2*theNumGenerators; i++)
            {
                //  Multiply the given matrix by the given generator
                //  to obtain theCandidate.
                MatrixMultiply( &theCandidate,
                                &theGenerators[i],
                                &theNode->mm);

                //  Is theCandidate's translation distance too great?
                switch (theSpaceType)
                {
                    case SpaceSpherical:
                        theTranslationDistance = acos(theCandidate.m[3][3]);
                        break;

                    case SpaceFlat:
                        theTranslationDistance = sqrt(
                                theCandidate.m[3][0]*theCandidate.m[3][0]
                              + theCandidate.m[3][1]*theCandidate.m[3][1]
                              + theCandidate.m[3][2]*theCandidate.m[3][2]);
                        break;

                    case SpaceHyperbolic:
                        theTranslationDistance = ArcCosh(theCandidate.m[3][3]);
                        break;
                }
                if (theTranslationDistance > theCutoffRadius)
                    continue;

                //  Is theCandidate already on the list?
                theCandidateIsOnTheList = FALSE;
                for (   theExistingNode = theFirstNode;
                        theExistingNode != NULL;
                        theExistingNode = theExistingNode->next)
                    if (SameMatrix(&theCandidate, &theExistingNode->mm) == TRUE)
                    {
                        theCandidateIsOnTheList = TRUE;
                        break;
                    }

                //  If the candidate isn't already on the list,
                //  add it at the end.
                if (theCandidateIsOnTheList == FALSE)
                {
                    theLastNode->next = new MatrixNode;
                    if (theLastNode->next == NULL)
                        throw new CException("Memory allocation failed.");
                    theLastNode->next->mm = theCandidate;
                    theLastNode->next->next = NULL;
                    theLastNode = theLastNode->next;
                    theNumMatrices++;
                }
            }
        }

        //  We're done with theGenerators.
        SAFE_DELETE_(theGenerators);

        //  Copy the matrices to an array.
        theMatrices = new Matrix[theNumMatrices];
        if (theMatrices == NULL)
            throw new CException("Memory allocation failed.");
        i = 0;
        for (   theNode = theFirstNode;
                theNode != NULL;
                theNode = theNode->next)
        {
            theMatrices[i++] = theNode->mm;
        }
        if (i != theNumMatrices)
            throw new CException("Internal error involving theNumMatrices.");

        //  Free the linked list.
        while (theFirstNode != NULL)
        {
            theSecondNode = theFirstNode->next;
            SAFE_DELETE(theFirstNode);
            theFirstNode = theSecondNode;
        }

        //  Allocate space for the parities.
        theParities = new ParityType[theNumMatrices];
        if (theParities == NULL)
            throw new CException("Memory allocation failed.");

        //  Compute the parity of each group element.
        for (i = 0; i < theNumMatrices; i++)
            theParities[i] =
                (MatrixDeterminant(&theMatrices[i]) > 0.0f ?
                    ParityNormal : ParityReversed);

        //  Mathematicians using this program may create
        //  their own spherical manifolds by combining
        //  a left-handed and a right-handed group.
        //  Let's do them a favor and check that the result
        //  contains no group element with fixed points
        //  except the identity.
        //  To test for fixed points it's enough to test whether
        //  1 is a root of the characteristic polynomial, which
        //  is equivalent to testing whether det(g - I) is zero.
        //
        //  Comment out these lines to allow for orbifolds.
#if 0
        if (theSpaceType == SpaceSpherical)
            for (i = 1; i < theNumMatrices; i++)
            {
                theMatrix = theMatrices[i];
                for (j = 0; j < 4; j++)
                    theMatrix.m[j][j] -= 1.0f;
                if (fabs(MatrixDeterminant(&theMatrix)) < SINGULAR_EPSILON)
                    throw new CException("Group contains fixed points.");
            }
#endif

        //  We've now finished all the stuff that might fail,
        //  such as reading the file and allocating memory.
        //  So let's drop out of the try{} block.
    }
    catch (CException *anException)
    {
        MessageBox(NULL, anException->Message(), "File Error", MB_OK);
        SAFE_DELETE(anException)

        if (theFile != NULL)
            fclose(theFile);
        SAFE_DELETE_(theGenerators);
        SAFE_DELETE_(theMatrices);
        SAFE_DELETE_(theParities);
        while (theFirstNode != NULL)
        {
            theSecondNode = theFirstNode->next;
            SAFE_DELETE(theFirstNode);
            theFirstNode = theSecondNode;
        }

        return;
    }

    //  Install the new matrices.
    SAFE_DELETE_(gGroupElements);
    SAFE_DELETE_(gGroupElementParities);
    gGroupElements          = theMatrices;
    gGroupElementParities   = theParities;
    gNumGroupElements       = theNumMatrices;
    gSpaceType              = theSpaceType;

    //  Recompute the projection matrices for the new SpaceType.
    InitProjectionMatrices();

    //  Reset the camera.
    MatrixIdentity(&gViewMatrix);
    gCameraParity   = ParityNormal;
//  gCameraSpeed    = 0.0f;     //  uncomment to reset camera speed

    //  Compute the new Dirichlet domain.
    SAFE_DELETE(gDirichletDomain);
    gDirichletDomain = new CDirichletDomain(
        gNumGroupElements,
        gGroupElements,
        (SpaceTypeDD) gSpaceType);  //  needed for consistent beam widths

    //  Set the file name as the window title,
    //  after skipping past the directory path and
    //  temporarily suppressing the file name extension.
    theWindowTitle = aFileName;
    for (theChar = aFileName; *theChar != 0; theChar++)
        if (*theChar == '\\')
            theWindowTitle = theChar + 1;
    theWindowTitle[strlen(theWindowTitle) - 4] = 0;
    SetWindowText(gWindow, theWindowTitle);
    theWindowTitle[strlen(theWindowTitle)] = '.';
}


static BOOL SameMatrix(
    Matrix  *a,
    Matrix  *b)
{
    int i,
        j;

    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            if (fabs(a->m[i][j] - b->m[i][j]) > MATRIX_EPSILON)
                return FALSE;

    return TRUE;
}


static void ReadMatrixFile(char *aFileName)
{
    int         theOldNumGroupElements;
    Matrix      *theOldGroupElements;
    ParityType  *theOldGroupElementParities;
    FILE        *theFile;
    float       theDummy;
    int         theCount,
                theResult,
                i,
                j,
                k;
    float       theHeight;
    char        *theWindowTitle,
                *theChar;

    //  Read the matrices from the given file.

    //  Keep the old matrices around until the new ones have been read.
    theOldNumGroupElements      = gNumGroupElements;
    theOldGroupElements         = gGroupElements;
    theOldGroupElementParities  = gGroupElementParities;

    gNumGroupElements       = 0;
    gGroupElements          = NULL;
    gGroupElementParities   = NULL;

    try
    {
        //  Open the file.
        theFile = fopen(aFileName, "r");
        if (theFile == NULL)
            throw new CException("Couldn't open file.");

        //  Count the total number of matrix entries in the file.
        theCount = 0;
        do
        {
            theResult = fscanf(theFile, "%f", &theDummy);
            if (theResult == 1)
                theCount++;
            else if (theResult != EOF)
                throw new CException("File contains non-numeric entries.");
        } while (theResult == 1);
        rewind(theFile);

        //  The number of entries should be a positive multiple of 16.
        if (theCount%16 != 0)
            throw new CException("File must contain 4x4 matrices.");
        if (theCount == 0)
            throw new CException("File is empty.");

        //  How many matrices does the file contain?
        gNumGroupElements = theCount/16;

        //  Allocate space for the new matrices.
        gGroupElements = new Matrix[gNumGroupElements];
        if (gGroupElements == NULL)
            throw new CException("Memory allocation failed.");

        //  Read the new matrices.
        for (i = 0; i < gNumGroupElements; i++)
        {
            for (j = 0; j < 4; j++)
                for (k = 0; k < 4; k++)
                    fscanf(theFile, "%f", &gGroupElements[i].m[j][k]);

            //  2002/10/13  Conjugate by the reflection diag(1,1,-1,1)
            //  to compensate for the difference in orientation conventions
            //  between OpenGL and Direct3D.
            for (j = 0; j < 4; j++)
            {
                gGroupElements[i].m[j][2] = - gGroupElements[i].m[j][2];
                gGroupElements[i].m[2][j] = - gGroupElements[i].m[2][j];
            }
        }

        //  Close the file.
        fclose(theFile);

        //  Compute the parity of each group element.
        gGroupElementParities = new ParityType[gNumGroupElements];
        if (gGroupElementParities == NULL)
            throw new CException("Memory allocation failed.");
        for (i = 0; i < gNumGroupElements; i++)
            gGroupElementParities[i] =
                (MatrixDeterminant(&gGroupElements[i]) > 0.0f ?
                    ParityNormal : ParityReversed);

        //  Delete the old group elements.
        SAFE_DELETE_(theOldGroupElements);
        SAFE_DELETE_(theOldGroupElementParities);
    }
    catch (CException *anException)
    {
        MessageBox(NULL, anException->Message(), "File Error", MB_OK);
        SAFE_DELETE(anException)

        if (theFile != NULL)
            fclose(theFile);

        SAFE_DELETE_(gGroupElements);
        SAFE_DELETE_(gGroupElementParities);

        gNumGroupElements       = theOldNumGroupElements;
        gGroupElements          = theOldGroupElements;
        gGroupElementParities   = theOldGroupElementParities;

        return;
    }

    //  Deduce the SpaceType from the matrices.
    //  If the first matrix is the identity, we may
    //  need to look at the second matrix to figure it out.

    if (gGroupElements[0].m[3][3] != 1.0f || gNumGroupElements == 1)
        theHeight = gGroupElements[0].m[3][3];
    else
        theHeight = gGroupElements[1].m[3][3];

    if      (theHeight <  1.0f)  gSpaceType = SpaceSpherical;
    else if (theHeight == 1.0f)  gSpaceType = SpaceFlat;
    else                         gSpaceType = SpaceHyperbolic;

    //  Recompute the projection matrices for the new SpaceType.
    InitProjectionMatrices();

    //  Reset the camera position, but maintain the previous speed.
    MatrixIdentity(&gViewMatrix);
    gCameraParity   = ParityNormal;
//  gCameraSpeed    = 0.0f;     //  uncomment to reset camera speed

    //  Compute the new Dirichlet domain.
    SAFE_DELETE(gDirichletDomain);
    gDirichletDomain = new CDirichletDomain(
        gNumGroupElements,
        gGroupElements,
        (SpaceTypeDD) gSpaceType);  //  needed for consistent beam widths

    //  Set the file name as the window title,
    //  after skipping past the directory path and
    //  temporarily suppressing the file name extension.
    theWindowTitle = aFileName;
    for (theChar = aFileName; *theChar != 0; theChar++)
        if (*theChar == '\\')
            theWindowTitle = theChar + 1;
    theWindowTitle[strlen(theWindowTitle) - 4] = 0;
    SetWindowText(gWindow, theWindowTitle);
    theWindowTitle[strlen(theWindowTitle)] = '.';
}


static void Update(void)
{
    //  Tell the camera and each scenery object how many
    //  seconds have elapsed since the last Update().

    static unsigned int theStartTime    = 0,
                        theNumFrames    = 0;
    static double       theFrameRate    = 30.0;
    static BOOL         theFirstSecond  = TRUE;
    
    unsigned int    theCurrentTime,
                    theElapsedTime;
    double          theNewFrameRate;
    char            theWindowTitle[128];
    Matrix          theCameraPosition;

    //  Get the time in milliseconds.
    theCurrentTime = GetTickCount();

    //  One-time initialization.
    if (theStartTime == 0)
    {
        theStartTime = theCurrentTime;
    }
    else
    {
        //  Count the number of frames drawn in approximately one second.
        theNumFrames++;
        theElapsedTime = theCurrentTime - theStartTime;
        if (theFirstSecond == TRUE && theElapsedTime > 0)
        {
            //  During the first second of operation,
            //  revise theFrameRate after each clock tick.
            //  At first it won't be super accurate, but at least
            //  it'll be better than our arbitrary initial guess.
            theFrameRate = theNumFrames / (0.001 * theElapsedTime);
        }
        if (theElapsedTime > 1000)
        {
            theFirstSecond = FALSE;

            theNewFrameRate = theNumFrames / (0.001 * theElapsedTime);
            theNumFrames = 0;
            theStartTime = theCurrentTime;
        
            //  If the user holds down a menu, our program gets suspended
            //  but the clock keeps running.  As a result, the nominal
            //  frame rate is too low and the scenery moves too fast
            //  in compensation.  To avoid this problem, don't let
            //  the nominal frame rate decrease too fast.
            if (theNewFrameRate < 0.8 * theFrameRate)
                theNewFrameRate = 0.8 * theFrameRate;
        
            theFrameRate = theNewFrameRate;

            if (gShowFrameRate)
            {
                wsprintf(theWindowTitle, "%3d frames per second", (int)theFrameRate);
                SetWindowText(gWindow, theWindowTitle);
            }
        }
    }

    //  Update the camera position, as embodied by gViewMatrix.
    //  In principle we want to move the camera a tiny bit forward.
    //  In practice we move the world a tiny bit backward in camera space.
    //  In OpenGL the camera looks down the negative z-axis,
    //  so "backward" means in the positive z-direction.
    AppendTranslation(&gViewMatrix, 2, gCameraSpeed / theFrameRate);

    //  If a Dirichlet domain is present, keep the camera within
    //  the central cell.  (Note:  This isn't necessary in the
    //  spherical case -- where the whole space is finite -- but
    //  does no harm there.)
    if (gDirichletDomain != NULL)
    {
        MatrixInverse(&theCameraPosition, &gViewMatrix);
        gDirichletDomain->CenterCamera(&theCameraPosition);
        MatrixInverse(&gViewMatrix, &theCameraPosition);
        gCameraParity = 
            (MatrixDeterminant(&gViewMatrix) > 0.0f ?
                ParityNormal : ParityReversed);
    }

    //  Normalize gViewMatrix to keep it legal for the current SpaceType.
    NormalizeViewMatrix();

    //  Update the SpinningEarth.
    gSpinningEarth->Update(1.0 / theFrameRate);
}


static void NormalizeViewMatrix(void)
{
    //  Numerical errors can accumulate and force gViewMatrix
    //  "out of round", in the sense that its rows are no
    //  longer orthonormal.  This effect is small in spherical and
    //  flat spaces, but can be significant in hyperbolic spaces,
    //  especially if the camera travels far from the origin.

    //  The "Gram-Schmidt process" consists of rescaling each
    //  row to restore unit length, and subtracting small multiples
    //  of one row from another to restore orthogonality.
    //  Here we carry out a first-order approximation to the
    //  Gram-Schmidt process.  That is, we normalize each row
    //  to unit length, but then assume that the subsequent
    //  orthogonalization step doesn't spoil the unit length.
    //  This assumption will be well satisfied because small
    //  first order changes orthogonal to a given vector affect
    //  its length only to second order.

    //  If your primary goal is to understand the techniques
    //  of curved space rendering, you may safely skip over
    //  the details of this orthonormalization. 


    //  Note:  OpenGL interprets m[i][*] as a column rather than a row,
    //  and considers matrices to act as (matrix)(column vector).
    //  Here we interpret m[i][*] as a row and consider matrices
    //  to act as (row vector)(matrix).  The two approaches are
    //  100% equivalent, because each is the transpose of the other.

    Matrix  m1,
            m2,
            theProduct;
    float   theFactor;
    int     i,
            j,
            k;

    //  Concoct a matrix giving the inner products
    //  of gViewMatrix's rows with one another.
    //  (This is a bit ad hoc, but it works.)

    m1 = gViewMatrix;
    MatrixTranspose(&m2, &m1);

    switch (gSpaceType)
    {
        case SpaceSpherical:
            //  no change required
            break;

        case SpaceFlat:

            //  zero out the mixed terms in both m1 and m2

            m1.m[3][0] = 0.0f;
            m1.m[3][1] = 0.0f;
            m1.m[3][2] = 0.0f;

            m2.m[0][3] = 0.0f;
            m2.m[1][3] = 0.0f;
            m2.m[2][3] = 0.0f;

            break;

        case SpaceHyperbolic:

            //  negate the mixed terms in m2 only

            m2.m[0][3] = - m2.m[0][3];
            m2.m[1][3] = - m2.m[1][3];
            m2.m[2][3] = - m2.m[2][3];

            m2.m[3][0] = - m2.m[3][0];
            m2.m[3][1] = - m2.m[3][1];
            m2.m[3][2] = - m2.m[3][2];

            break;
    }

    MatrixMultiply(&theProduct, &m1, &m2);

    //  Normalize the length of each row of gViewMatrix.
    for (i = 0; i < 4; i++)
    {
        if (theProduct.m[i][i] <= 0.0f)
            exit(1);

        theFactor = (float) (1.0 / sqrt(theProduct.m[i][i]));

        for (j = 0; j < 4; j++)
            gViewMatrix.m[i][j] *= theFactor;
    }

    //  Make the rows of gViewMatrix mutually orthogonal.
    for (i = 4; --i >= 0; )         // iterate backwards
        for (j = i; --j >= 0; )     // iterate backwards
        {
            theFactor = theProduct.m[j][i];

            for (k = 0; k < 4; k++)
                gViewMatrix.m[j][k] -=
                    theFactor * gViewMatrix.m[i][k];
                    
        }
}


static void Render(void)
{
    if (gWindow == NULL
     || IsIconic(gWindow))
        return;

    //  Enable depth testing.
    glEnable(GL_DEPTH_TEST);

    //  Enable textures.
    glEnable(GL_TEXTURE_2D);

    //  Request perspectively correct textures.
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

    //  We won't use explicit lights.
    glDisable(GL_LIGHTING);

    //  Enable backface culling.
    glEnable(GL_CULL_FACE);

    //  OpenGL doesn't support explicit fogging,
    //  so instead fade the scene to black.
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, gFog ? GL_MODULATE : GL_REPLACE);

    //  As explained in comment (3) at the top of this file,
    //  OpenGL implementations conforming to the official OpenGL ARB standard
    //  will automatically clip to z < 0.  In the (unlikely?) event that
    //  your system is non-compliant, uncomment the following lines of code
    //  to enable an explicit z < 0 clipping plane (in eye coordinates).
//  if (gSpaceType == SpaceSpherical)
//  {
//      static const GLdouble   theFrontHemispherePlane[4] = {0, 0, -1, 0};
//
//      glEnable(GL_CLIP_PLANE0);
//      glMatrixMode(GL_MODELVIEW);
//      glLoadIdentity();
//      glClipPlane(GL_CLIP_PLANE0, theFrontHemispherePlane);
//  }
//  else
//      glDisable(GL_CLIP_PLANE0);

    //  Initialize the color and depth buffers.
    if (gSpaceType == SpaceSpherical
     && gStereo == FALSE)
    {
        glDisable(GL_DEPTH_TEST);
        DrawSpaceshipBackground();
        glEnable(GL_DEPTH_TEST);
        glClear(GL_DEPTH_BUFFER_BIT);
    }
    else
    {
        //  Clear the color and depth buffers.
        glClearColor(0,0,0,1);  //  Clear to black to match fog.
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    }

    //  Position the camera.
    glMatrixMode(GL_MODELVIEW);
    glLoadMatrixf((float *) &gViewMatrix.m);

    //  Draw.
    if (gStereo == FALSE)
    {
        //  Draw a full color image for a single eye.
        ProjectAndDraw(EyeOnly);
    }
    else
    {
        //  Restrict to the red channel.
        glColorMask(TRUE, FALSE, FALSE, TRUE);

        //  Draw the left eye image.
        ProjectAndDraw(EyeLeft);

        //  Clear the z-buffer.
        glClear(GL_DEPTH_BUFFER_BIT);

        //  Restrict to the "cyan channel".
        glColorMask(FALSE, TRUE, TRUE, TRUE);

        //  Draw the right eye image.
        ProjectAndDraw(EyeRight);

        //  Re-enable all color channels.
        glColorMask(TRUE, TRUE, TRUE, TRUE);
    }

    glFlush();

    SwapBuffers(gDeviceContext);
}


static void DrawSpaceshipBackground(void)
{
    //  In a spherical space we see ourselves in all directions,
    //  so the back of our own spaceship fills the sky.
    //  No matter where we move, the image of our own spaceship
    //  is always the same, so draw a precomputed image instead
    //  of rendering the spaceship in real time.

    if (gFog)
        glColor3f(SPHERICAL_BACKGROUND_FOG, SPHERICAL_BACKGROUND_FOG, SPHERICAL_BACKGROUND_FOG);
    else
        glColor3f(1, 1, 1); //  no fog

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glBindTexture(GL_TEXTURE_2D, TextureSpaceship + 1);

    glCullFace(GL_BACK);

    glBegin(GL_QUADS);
        glTexCoord2f(0, 1);  glVertex2f(-1, -1);
        glTexCoord2f(1, 1);  glVertex2f( 1, -1);
        glTexCoord2f(1, 0);  glVertex2f( 1,  1);
        glTexCoord2f(0, 0);  glVertex2f(-1,  1);
    glEnd();
}


static void ProjectAndDraw(
    EyeType   anEyeType)
{
    //  Standard rendering shows all visible scenery in flat and hyperbolic spaces.
    //  In a spherical space it shows only the front hemisphere, but if
    //  the scenery objects are distributed symmetrically (with each object
    //  having a twin at its antipodal point) nothing is lost by omitting
    //  the back hemisphere.

    if (gSpaceType != SpaceSpherical
     || gRenderBackHemisphere == FALSE)
    {
        //  Draw the scene.
        DrawContents(anEyeType, BananaFull);
    }
    else
    {
        //  Render the front hemisphere viewing banana
        //  into the front half of the clipping box (-1 <= z <= 0).
        DrawContents(anEyeType, BananaFront);

        //  Render the back hemisphere viewing banana
        //  into the back half of the clipping box (0 <= z <= +1).
        DrawContents(anEyeType, BananaBack);
    }
}


static void DrawContents(
    EyeType     anEyeType,
    RangeType   aRangeType)
{
    Matrix  theModelviewMatrix;
    float   theFogFactor,
            theDistance;
    int     i;

    static GLfloat  theOrigin[4] = {0, 0, 0, 1};
 
    //  Set the projection matrix.
    glMatrixMode(GL_PROJECTION);
    glLoadMatrixf((float *) &gProjectionMatrix[anEyeType][aRangeType].m);

    //  If we were using built-in lighting we could package the following calls
    //  into a display list and ship them off to the graphics card once and for all.
    //  But because we are doing our own (primitive) lighting with the fog factor,
    //  we have to execute this loop in real-time to properly alternate
    //  the glColor3f() calls with the scenery rendering.

    glMatrixMode(GL_MODELVIEW);

    for (i = 0; i < gNumGroupElements; i++)
    {
        //  Cull back-facing triangles.
        glCullFace((gGroupElementParities[i] == gCameraParity) ? GL_BACK : GL_FRONT);

        //  Push the i-th group element onto the modelview stack.
        glPushMatrix();
        glMultMatrixf((float *) &gGroupElements[i].m);

        //  Set a single fog factor for this cell.
        //  (For smoother fogging we could compute a per-vertex fog factor,
        //  as in the Direct3D version of this program.)
        //  Note that we need to call ComputeFogFactor() whether
        //  fog is on or not, because we need theDistance.
        glGetFloatv(GL_MODELVIEW_MATRIX, (float *)theModelviewMatrix.m);
        theFogFactor = ComputeFogFactor(theModelviewMatrix.m[3], theOrigin, aRangeType, &theDistance);
        if (!gFog)
            theFogFactor = 1;

        //  Draw the Earth (and any other scenery the programmer may add).
        if (gViewEarth)
            gSpinningEarth->RenderDisplayList(theFogFactor, theDistance);

        //  Draw the framework of the Dirichlet domain.
        //  [Putting the beams into a display list didn't help.
        //  In fact it might have slowed the program down a tiny bit
        //  on Clarence.]  [Carla takes about 22 ms to draw beams only
        //  (without display lists) or 14 ms to draw Earths only
        //  (with display lists) or 38 ms to draw both, all for
        //  the hyperbolic tiling Dodec-Oct.deep.gen, so it seems
        //  unlikely that converting to display lists will make
        //  a big difference.]
        if (gViewBeams)
            gDirichletDomain->RenderBeams(theFogFactor, TextureBeam + 1);

        glPopMatrix();
    }

    //  Restore the default cull mode.
    glCullFace(GL_BACK);
}


static float ComputeFogFactor(
    float   *u,
    float   *v,
    int     aRangeType, //  really a RangeType, but cast as an int
                        //      so we can pass the parameter through
                        //      external code
    float   *aDistance)
{
    //  This code is designed to produce an easily viewable result.
    //  However, in a "real hypersphere", for example, the curvature
    //  of the space would tend to refocus the light from distant objects,
    //  offsetting the fog effect and possibly making distant objects
    //  brighter than intermediate objects of the same intrinsic luminosity.

    float   theDistance,
            theFogFactor;

    switch (gSpaceType)
    {
        case SpaceSpherical:
            theDistance = (float) (acos(  u[0]*v[0]
                                        + u[1]*v[1]
                                        + u[2]*v[2]
                                        + u[3]*v[3] ));
            switch ((RangeType) aRangeType)
            {
                case BananaFull:
                    theFogFactor = (theDistance / PI) * SPHERICAL_FOG_AT_ANTIPODE;
                    break;
                case BananaFront:
                    theFogFactor = (theDistance / PI) * SPHERICAL_FOG_AT_ANTIPODE;
                    break;
                case BananaBack:
                    //  Keep in mind here that theDistance is measured
                    //  from the antipodal camera position, not the
                    //  original camera position.
                    theFogFactor = SPHERICAL_FOG_AT_ANTIPODE
                          +  (SPHERICAL_FOG_MAX - SPHERICAL_FOG_AT_ANTIPODE)
                             * (1.0f - (theDistance / PI));
                    break;
            }
            break;

        case SpaceFlat:
            theDistance = (float) (sqrt(  (u[0] - v[0])*(u[0] - v[0])
                                        + (u[1] - v[1])*(u[1] - v[1])
                                        + (u[2] - v[2])*(u[2] - v[2]) ));
            if (theDistance <= FLAT_FOG_BEGIN)
                theFogFactor = 0.0f;
            else
            if (theDistance >= FLAT_FOG_END)
                theFogFactor = 1.0f;
            else
                theFogFactor = (theDistance  - FLAT_FOG_BEGIN)
                             / (FLAT_FOG_END - FLAT_FOG_BEGIN);
            break;

        case SpaceHyperbolic:
            theDistance = (float) (ArcCosh( - u[0]*v[0]
                                            - u[1]*v[1]
                                            - u[2]*v[2]
                                            + u[3]*v[3] ));
            if (theDistance <= HYPERBOLIC_FOG_BEGIN)
                theFogFactor = 0.0f;
            else
            if (theDistance >= HYPERBOLIC_FOG_END)
                theFogFactor = 1.0f;
            else
                theFogFactor = (   theDistance     - HYPERBOLIC_FOG_BEGIN)
                             / (HYPERBOLIC_FOG_END - HYPERBOLIC_FOG_BEGIN);
            break;
    }

    if (aDistance != NULL)
    {
        //  Report the distance to the caller.
        *aDistance = theDistance;

        //  In the spherical case, what counts is
        //  the distance from the camera or its antipodal
        //  point, whichever is less.
        if (gSpaceType == SpaceSpherical
         && *aDistance > PI/2)
            *aDistance = PI - *aDistance;
    }

    return 1.0f - theFogFactor;
}

static double ArcCosh(double x)
{
    if (x >= 1.0)
        return log(x + sqrt(x*x - 1.0));
    else
        return 0.0;
}


static void MatrixTranslation(
    Matrix  *aMatrix,
    int     anAxis,
    double  aDistance)
{
    //  Start with an identity matrix and then...
    MatrixIdentity(aMatrix);

    //  ...splice in the SpaceType-dependent translational part.
    switch (gSpaceType)
    {
        case SpaceSpherical:
            aMatrix->m[anAxis][anAxis]    = (float)  cos(aDistance);
            aMatrix->m[3][3]              = (float)  cos(aDistance);
            aMatrix->m[anAxis][3]         = (float) -sin(aDistance);
            aMatrix->m[3][anAxis]         = (float)  sin(aDistance);
            break;

        case SpaceFlat:
            aMatrix->m[3][anAxis]         = (float) aDistance;
            break;

        case SpaceHyperbolic:
            aMatrix->m[anAxis][anAxis]    = (float) cosh(aDistance);
            aMatrix->m[3][3]              = (float) cosh(aDistance);
            aMatrix->m[anAxis][3]         = (float) sinh(aDistance);
            aMatrix->m[3][anAxis]         = (float) sinh(aDistance);
            break;
    }
}


static void PrependTranslation(
    Matrix  *aMatrix,
    int     anAxis,
    double  aDistance)
{
    Matrix  temp;

    MatrixTranslation(&temp, anAxis, aDistance);
    MatrixMultiply(aMatrix, &temp, aMatrix);
}


static void AppendTranslation(
    Matrix  *aMatrix,
    int     anAxis,
    double  aDistance)
{
    Matrix  temp;

    MatrixTranslation(&temp, anAxis, aDistance);
    MatrixMultiply(aMatrix, aMatrix, &temp);
}


static void AppendRotation(
    Matrix  *aMatrix,
    float   aPitch,
    float   aYaw,
    float   aRoll)
{
    Matrix  theRotation;

    MatrixRotation(&theRotation, 0, aPitch);
    MatrixMultiply(aMatrix, aMatrix, &theRotation);

    MatrixRotation(&theRotation, 1, aYaw);
    MatrixMultiply(aMatrix, aMatrix, &theRotation);

    MatrixRotation(&theRotation, 2, aRoll);
    MatrixMultiply(aMatrix, aMatrix, &theRotation);
}
