//  SpinningEarth-GL.cpp
//
//       YOU DO NOT NEED TO READ THIS FILE TO UNDERSTAND
//             COMPUTER GRAPHICS IN CURVED SPACES.
//
//  CurvedSpaces-Simple includes the much simpler SceneryObjectCube
//  and SceneryObjectOctahedron, which illustrate the basic principles.
//
//  2003/12/20  Draw extremely distant images as points.
//
//  2003/12/20  Comment:
//  We could in principle use an octahedral triangulation
//  at low resolution and an icosahedral one at high resolution.
//  For example, the progression could be
//
//       octahedron (0)       8 triangles
//       octahedron (1)      32 triangles
//      icosahedron (1)      80 triangles
//      icosahedron (2)     320 triangles
//
//  We can't use an icosahedron at low resolution because
//  (1) some of its faces straddle the equator,
//      creating problems with the northern/southern texture scheme,
//  (2) it has more faces than we really need at minimal resolution.
//
//  However, in practice the subdivided octahedron looks fine,
//  and the Earths aren't the bottleneck anyhow (the beams are)
//  especially now that distant Earths are drawn as dots.
//
//  2003/12/19  Consolidate parallel arrays into a single
//  array of structures to represent the various subdivisions
//  of the triangulation.
//
//  2003/12/19  Replace longitude-latitude texture parameterization
//  with separate northern hemisphere and southern hemisphere
//  projections.  Each hemisphere projects radially onto
//  an octahedron and then orthogonally onto the xy-plane.
//  This avoids the anomalies at the north and south poles
//  as well as the problem with coordinate-wrapping.
//  Note too that the texture coordinates agree at the equator,
//  so no repetition of vertices is needed.  Conceptually
//  the sphere is being mapped onto the texture plane
//  with no rips or tears -- the only distinction between
//  the northern and southern hemispheres is which texture
//  gets applied.
//
//  2003/01/10  Implement display lists to improve performance.
//
//  2002/10/13  Replace icosahedron having no duplicated vertices
//  with an octahedron having duplicated vertices, to accomodate
//  the fact that OpenGL cannot wrap texture coordinates properly.
//  (That is, when OpenGL goes from (0.50, 0.99) to (0.50, 0.01)
//  is always backtracks "the long way" rather than wrapping around
//  "the short way".)
//
//  2002/09/24  Mirror-reverses the Earth, whose triangulation was
//  originally laid out in Direct3D's left-handed coordinate system,
//  to accomodate OpenGL's right-handed coordinate system.

#include "SpinningEarth-GL.hpp"
#include <assert.h>
#include <math.h>

#define DEGREES_PER_SECOND  45
#define EARTH_RADIUS        0.05
#define PI                  3.14159265358979323846


//  The standard gluSphere() is easy to use,
//  but its rectangular triangulation scheme, aligned along
//  latitude and longitude lines, is woefully inefficient,
//  especially near the poles.  Use an efficient octahedral
//  or icosahedral triangulation instead, subdivided as necessary.


CSpinningEarth::CSpinningEarth(void)
{
    int i;

    itsRotationAngle = 0.0;

    for (i = 0; i <= REFINEMENT_LEVEL; i++)
    {
        itsSubdivisions[i].itsNumVertices           = 0;
        itsSubdivisions[i].itsVertices              = NULL;
        itsSubdivisions[i].itsTextureCoordinates    = NULL;
        itsSubdivisions[i].itsNumFaces              = 0;
        itsSubdivisions[i].itsFaces                 = NULL;
        itsSubdivisions[i].itsDisplayList           = 0;
    }

    InitOctahedron(&itsSubdivisions[0]);

    for (i = 0; i < REFINEMENT_LEVEL; i++)
        SubdivideTriangulation(&itsSubdivisions[i], &itsSubdivisions[i+1]);
}

CSpinningEarth::~CSpinningEarth(void)
{
    int i;

    for (i = 0; i <= REFINEMENT_LEVEL; i++)
    {
        SAFE_DELETE_(itsSubdivisions[i].itsVertices)
        SAFE_DELETE_(itsSubdivisions[i].itsTextureCoordinates)
        SAFE_DELETE_(itsSubdivisions[i].itsFaces)
    }

}


void CSpinningEarth::Update(double anElapsedTime)
{
    itsRotationAngle += DEGREES_PER_SECOND * anElapsedTime;

    if (itsRotationAngle >= 360.0)
        itsRotationAngle -= 360.0;
}


void CSpinningEarth::MakeDisplayLists(
    GLuint  aNorthernTexture,
    GLuint  aSouthernTexture)
{
    int             i;
    Triangulation   *s;

    for (i = 0; i <= REFINEMENT_LEVEL; i++)
    {
        s = &itsSubdivisions[i];

        s->itsDisplayList = glGenLists(1);
        if (s->itsDisplayList != 0)
        {
            glNewList(s->itsDisplayList, GL_COMPILE);

                glEnableClientState(GL_VERTEX_ARRAY);
                glEnableClientState(GL_TEXTURE_COORD_ARRAY);

                glVertexPointer  (3, GL_DOUBLE, 0, s->itsVertices          );
                glTexCoordPointer(2, GL_DOUBLE, 0, s->itsTextureCoordinates);

//              If OpenGL 1.2 is available, glDrawRangeElements() may be better than glDrawElements().
//              glDrawRangeElements(GL_TRIANGLES, 0, itsNumVertices[i] - 1, 3*itsNumFaces[i], GL_UNSIGNED_INT, itsFaces[i]);

                glBindTexture(GL_TEXTURE_2D, aNorthernTexture + 1);
                glDrawElements(GL_TRIANGLES, 3*s->itsNumFaces/2, GL_UNSIGNED_INT, s->itsFaces);

                glBindTexture(GL_TEXTURE_2D, aSouthernTexture + 1);
                glDrawElements(GL_TRIANGLES, 3*s->itsNumFaces/2, GL_UNSIGNED_INT, s->itsFaces + 3*s->itsNumFaces/2);

                glDisableClientState(GL_VERTEX_ARRAY);
                glDisableClientState(GL_TEXTURE_COORD_ARRAY);

            glEndList();
        }
    }
}


void CSpinningEarth::FreeDisplayLists(void)
{
    int i;

    for (i = 0; i <= REFINEMENT_LEVEL; i++)
    {
        glDeleteLists(itsSubdivisions[i].itsDisplayList, 1);
        itsSubdivisions[i].itsDisplayList = 0;
    }
}


void CSpinningEarth::RenderDisplayList(
    float   aFogFactor,
    float   aDistance)
{
    assert(REFINEMENT_LEVEL >= 3);

    //  Pick an appropriate level of detail at which
    //  to render this image of the spinning Earth.
    if (aDistance < 2.5)
    {
        glColor3f(aFogFactor, aFogFactor, aFogFactor);

        glPushMatrix();
        glRotated(-90, 1, 0, 0);
        glRotated(itsRotationAngle, 0, 0, 1);

        if (aDistance < 1.0)
            glCallList(itsSubdivisions[3].itsDisplayList);
        else
        if (aDistance < 1.58)   //  just above pi/2
            glCallList(itsSubdivisions[2].itsDisplayList);
        else
        if (aDistance < 2.0)
            glCallList(itsSubdivisions[1].itsDisplayList);
        else    //  aDistance < 2.5
            glCallList(itsSubdivisions[0].itsDisplayList);

        glPopMatrix();
    }
    else
    {
        if (aDistance < 3.0)
            glPointSize(3);
        else
        if (aDistance < 3.5)
            glPointSize(2);
        else
            glPointSize(1);
        glBegin(GL_POINTS);
        glColor3f(aFogFactor, aFogFactor, aFogFactor);
        glVertex3f(0,0,0);
        glEnd();
    }

}


//  The remaining methods implement the octahedral triangulation
//  of the sphere.  Feel free to ignore them.

void CSpinningEarth::InitOctahedron(
    Triangulation   *aTriangulation)
{
    int     i;

    static Vertex   v[6] =
                    {
                        { 1.0,  0.0,  0.0}, //  equator
                        { 0.0,  1.0,  0.0},
                        {-1.0,  0.0,  0.0},
                        { 0.0, -1.0,  0.0},
                        { 0.0,  0.0,  1.0}, //  north pole
                        { 0.0,  0.0, -1.0}  //  south pole
                    };

    //  Make sure all northern faces come before all southern faces,
    //  so they may be eventually be rendered separately using
    //  different texture maps.
    static GLuint   f[24] =
                    {
                        //  northern faces
                        4, 0, 1,
                        4, 1, 2,
                        4, 2, 3,
                        4, 3, 0,

                        //  southern faces
                        5, 1, 0,
                        5, 2, 1,
                        5, 3, 2,
                        5, 0, 3
                    };

    aTriangulation->itsNumVertices          = 6;
    aTriangulation->itsNumFaces             = 8;
    aTriangulation->itsVertices             = new Vertex[6];
    aTriangulation->itsTextureCoordinates   = new TextureUV[6];
    aTriangulation->itsFaces                = new GLuint[3*8];

    assert(aTriangulation->itsVertices              != NULL);
    assert(aTriangulation->itsTextureCoordinates    != NULL);
    assert(aTriangulation->itsFaces                 != NULL);

    for (i = 0; i < 6; i++)
    {
        aTriangulation->itsVertices[i][0] = v[i][0] * EARTH_RADIUS;
        aTriangulation->itsVertices[i][1] = v[i][1] * EARTH_RADIUS;
        aTriangulation->itsVertices[i][2] = v[i][2] * EARTH_RADIUS;

//      itsNormals[0][i][0] = v[i][0];
//      itsNormals[0][i][1] = v[i][1];
//      itsNormals[0][i][2] = v[i][2];

        ComputeTextureCoordinates(v[i], aTriangulation->itsTextureCoordinates[i]);
    }

    for (i = 0; i < 24; i++)
        aTriangulation->itsFaces[i] = f[i];
}

void CSpinningEarth::SubdivideTriangulation(
    Triangulation *aTriangulation,
    Triangulation *aSubdivision)
{
    //  Subdivide the triangulation, replacing
    //  each old triangle with four new ones.
    //
    //    /\
    //   /__\
    //  /_\/_\
    //

    GLuint      u,
                v,
                *f,
                ff[3];
    double      normal[3],
                theLengthSquared,
                theFactor;
    int         *theTable,
                theVertexCount,
                i,
                j,
                k;

    aSubdivision->itsNumVertices        = aTriangulation->itsNumVertices + 3*aTriangulation->itsNumFaces/2;
    aSubdivision->itsNumFaces           = 4 * aTriangulation->itsNumFaces;
    aSubdivision->itsVertices           = new Vertex[aSubdivision->itsNumVertices];
    aSubdivision->itsTextureCoordinates = new TextureUV[aSubdivision->itsNumVertices];
    aSubdivision->itsFaces              = new GLuint[3*aSubdivision->itsNumFaces];

    assert(aSubdivision->itsVertices           != NULL);
    assert(aSubdivision->itsTextureCoordinates != NULL);
    assert(aSubdivision->itsFaces              != NULL);

    //  Copy the vertices from the previous level.
    for (i = 0; i < aTriangulation->itsNumVertices; i++)
    {
        for (j = 0; j < 3; j++)
            aSubdivision->itsVertices[i][j] = aTriangulation->itsVertices[i][j];
        for (j = 0; j < 2; j++)
            aSubdivision->itsTextureCoordinates[i][j] = aTriangulation->itsTextureCoordinates[i][j];
    }
    theVertexCount = aTriangulation->itsNumVertices;

    //  Create one new vertex on each edge.
    //  Make a table to index them, so the two triangles sharing
    //  an edge can share the same vertex.  (The memory for the
    //  table grows as the square of the number of old vertices.
    //  For modest triangulations this won't be a problem.
    //  For large triangulations a fancier algorithm could be used.)

    //  In spirit theTable is a 2-dimensional array.
    //  But such an array is a hassle to allocate in C/C++,
    //  so instead we just replace theTable[i][j] with
    //  theTable[i*aTriangulation->itsNumVertices + j].
    theTable = new int[aTriangulation->itsNumVertices * aTriangulation->itsNumVertices];
    assert(theTable != NULL);

    for (i = 0; i < aTriangulation->itsNumVertices; i++)
        for (j = 0; j < aTriangulation->itsNumVertices; j++)
            theTable[i*aTriangulation->itsNumVertices + j] = -1;

    for (i = 0; i < aTriangulation->itsNumFaces; i++)
        for (j = 0; j < 3; j++)
        {
            u = aTriangulation->itsFaces[3*i +   j    ];
            v = aTriangulation->itsFaces[3*i + (j+1)%3];

            if (theTable[u*aTriangulation->itsNumVertices + v] == -1)
            {
                assert(theVertexCount < aSubdivision->itsNumVertices);

                theLengthSquared = 0.0;
                for (k = 0; k < 3; k++)
                {
                    normal[k] = (aTriangulation->itsVertices[u][k] + aTriangulation->itsVertices[v][k])/2;
                    theLengthSquared += normal[k]*normal[k];
                }
                theFactor = 1.0 / sqrt(theLengthSquared);
                for (k = 0; k < 3; k++)
                    normal[k] *= theFactor;
                //  This program doesn't use explicit lighting,
                //  but if desired you could record the normal vector
                //  at this point.
                
                for (k = 0; k < 3; k++)
                    aSubdivision->itsVertices[theVertexCount][k] = normal[k] * EARTH_RADIUS;

                ComputeTextureCoordinates(normal, aSubdivision->itsTextureCoordinates[theVertexCount]);

                theTable[u*aTriangulation->itsNumVertices + v] = theVertexCount;
                theTable[v*aTriangulation->itsNumVertices + u] = theVertexCount;

                theVertexCount++;
            }
        }
    assert(theVertexCount == aSubdivision->itsNumVertices);

    //  Create the new faces.
    //  If all the old northern faces precede
    //      all the old southern faces in aTriangulation,
    //  then all the new northern faces will precede
    //      all the new southern faces in aSubdivision.
    for (i = 0; i < aTriangulation->itsNumFaces; i++)
    {
        //  The old vertices incident to this face will be f[0], f[1] and f[2].
        f = aTriangulation->itsFaces + 3*i;

        //  The new ones will be ff[0], ff[1], and ff[2].
        //  Each ff[j] lies opposite the corresponding f[j].
        for (j = 0; j < 3; j++)
            ff[j] = theTable[f[(j+1)%3]*aTriangulation->itsNumVertices + f[(j+2)%3]];

        //  Create the new faces.

        aSubdivision->itsFaces[12*i +  0] = ff[0];
        aSubdivision->itsFaces[12*i +  1] = ff[1];
        aSubdivision->itsFaces[12*i +  2] = ff[2];

        aSubdivision->itsFaces[12*i +  3] = f[0];
        aSubdivision->itsFaces[12*i +  4] = ff[2];
        aSubdivision->itsFaces[12*i +  5] = ff[1];

        aSubdivision->itsFaces[12*i +  6] = f[1];
        aSubdivision->itsFaces[12*i +  7] = ff[0];
        aSubdivision->itsFaces[12*i +  8] = ff[2];

        aSubdivision->itsFaces[12*i +  9] = f[2];
        aSubdivision->itsFaces[12*i + 10] = ff[1];
        aSubdivision->itsFaces[12*i + 11] = ff[0];
    }

    SAFE_DELETE_(theTable)
}

void CSpinningEarth::ComputeTextureCoordinates(
    Vertex      v,
    TextureUV   t)
{
    double  x,
            y,
            z,
            theLength;

    //  Copy the vertex coordinates.
    x = v[0];
    y = v[1];
    z = v[2];

    //  Project radially onto an octahedron.
    theLength = fabs(x) + fabs(y) + fabs(z);
    if (theLength > 0)
    {
        x /= theLength;
        y /= theLength;
        z /= theLength;
    }
    else
    {
        //  Should never occur!
        t[0] = t[1] = 0.0;
        return;
    }

    //  Project orthogonally onto the xy-plane.
    //  (Just ignore z!)

    //  Rotate 45 degrees and rescale to [0,1].
    t[0] = 0.5*(x + y) + 0.5;
    t[1] = 0.5*(x - y) + 0.5;
    
    //  For insurance against rounding errors...
    if (t[0] < 0.0) t[0] = 0.0;
    if (t[0] > 1.0) t[0] = 1.0;
    if (t[1] < 0.0) t[1] = 0.0;
    if (t[1] > 1.0) t[1] = 1.0;
}
